from collections import OrderedDict
from copy import deepcopy
from datetime import date
from datetime import datetime
from requests import Response
from typing import Dict
from typing import Optional
from unittest import mock
from unittest import TestCase
from unittest.mock import call
from unittest.mock import MagicMock
from uuid import uuid4

import pytest
from os2mo_helpers.mora_helpers import MoraHelper

from .test_config import DEFAULT_CHANGED_AT_SETTINGS
from sdlon.config import ChangedAtSettings
from sdlon.fix_departments import FixDepartments


def mock_sd_lookup(service_name, expected_params, response):
    base_responses = {
        "GetDepartment20111201": {
            "Department": [
                {
                    "DepartmentLevelIdentifier": _TestableFixDepartments.MO_CLASS_USER_KEY,
                    "DepartmentIdentifier": _TestableFixDepartments.SD_DEPARTMENT_SHORTNAME,
                    "DepartmentName": _TestableFixDepartments.SD_DEPARTMENT_NAME,
                    "DepartmentUUIDIdentifier": "99999999-9999-9999-9999-999999999999",
                    "ActivationDate": "2019-01-01",
                    "DeactivationDate": "9999-12-31",
                }
            ],
        }
    }

    _response = deepcopy(base_responses[service_name])
    _response.update(response)
    sd_lookup_path = "sdlon.fix_departments.sd_lookup"
    sd_lookup_mock = mock.patch(sd_lookup_path, return_value=_response)
    return sd_lookup_mock


class _TestableFixDepartments(FixDepartments):
    MO_ORG_ROOT = "00000000-0000-0000-0000-000000000000"
    MO_CLASS_USER_KEY = "Enhed"
    MO_CLASS_UUID = str(uuid4())
    SD_INSTITUTION_UUID = str(uuid4())
    SD_DEPARTMENT_NAME = "some department name"
    SD_DEPARTMENT_SHORTNAME = "some department short name"
    SD_DEPARTMENT_PARENT_UUID = str(uuid4())

    @classmethod
    def get_instance(cls, settings_dict: Optional[Dict] = None):
        all_settings_dict = deepcopy(DEFAULT_CHANGED_AT_SETTINGS)
        if settings_dict is not None:
            all_settings_dict.update(settings_dict)
        settings = ChangedAtSettings.parse_obj(all_settings_dict)

        read_mo_org = "sdlon.fix_departments.MoraHelper.read_organisation"
        with mock.patch(read_mo_org, return_value=cls.MO_ORG_ROOT):
            return cls(settings)

    def get_institution(self):
        return self.SD_INSTITUTION_UUID

    def get_parent(self, unit_uuid, validity_date):
        return self.SD_DEPARTMENT_PARENT_UUID

    def _get_mora_helper(self, settings):
        mock_helper = mock.MagicMock(spec=MoraHelper)
        mock_helper.read_organisation = mock.Mock(
            return_value=_TestableFixDepartments.MO_ORG_ROOT
        )
        mock_helper.read_classes_in_facet = mock.Mock(
            return_value=[
                [{"user_key": self.MO_CLASS_USER_KEY, "uuid": self.MO_CLASS_UUID}]
            ]
        )
        return mock_helper


class TestFixDepartmentsRootSetting:
    alternate_root = str(uuid4())

    @pytest.mark.parametrize(
        "settings,expected_root",
        [
            # Case 1: Default root
            (
                {},
                _TestableFixDepartments.MO_ORG_ROOT,
            ),
            # Case 2: Alternate root
            (
                {
                    "sd_fix_departments_root": alternate_root,
                },
                alternate_root,
            ),
        ],
    )
    def test_root(self, settings, expected_root):
        instance = _TestableFixDepartments.get_instance(settings_dict=settings)
        assert instance.org_uuid == expected_root


class TestFixDepartment(TestCase):
    def test_update_root_unit(self):
        instance = _TestableFixDepartments.get_instance()
        instance.get_parent = MagicMock(return_value=None)
        instance._create_org_unit_if_missing_in_mo = MagicMock(return_value=False)
        with mock_sd_lookup("GetDepartment20111201", {}, {}):
            instance.fix_department(
                "99999999-9999-9999-9999-999999999999", date(2020, 1, 1)
            )
            instance.helper._mo_post.assert_called_with(
                "details/edit",
                {
                    "type": "org_unit",
                    "data": {
                        "uuid": "99999999-9999-9999-9999-999999999999",
                        "user_key": instance.SD_DEPARTMENT_SHORTNAME,
                        "name": instance.SD_DEPARTMENT_NAME,
                        "parent": {"uuid": instance.MO_ORG_ROOT},
                        "org_unit_level": {"uuid": instance.MO_CLASS_UUID},
                        "org_unit_type": {"uuid": instance.MO_CLASS_UUID},
                        "validity": {"from": "2019-01-01", "to": None},
                    },
                },
            )

    def test_multiple_sd_root_department_registrations(self):
        instance = _TestableFixDepartments.get_instance()
        instance.get_parent = MagicMock(return_value=None)
        instance._create_org_unit_if_missing_in_mo = MagicMock(return_value=False)
        sd_response = {
            "Department": [
                {
                    "DepartmentLevelIdentifier": _TestableFixDepartments.MO_CLASS_USER_KEY,
                    "DepartmentIdentifier": _TestableFixDepartments.SD_DEPARTMENT_SHORTNAME,
                    "DepartmentUUIDIdentifier": "uuid",
                    "DepartmentName": _TestableFixDepartments.SD_DEPARTMENT_NAME,
                    "ActivationDate": "2019-01-01",
                    "DeactivationDate": "2023-12-31",
                },
                {
                    "DepartmentLevelIdentifier": _TestableFixDepartments.MO_CLASS_USER_KEY,
                    "DepartmentIdentifier": _TestableFixDepartments.SD_DEPARTMENT_SHORTNAME,
                    "DepartmentUUIDIdentifier": "uuid",
                    "DepartmentName": "new name",
                    "ActivationDate": "2024-01-01",
                    "DeactivationDate": "9999-12-31",
                },
            ]
        }
        with mock_sd_lookup("GetDepartment20111201", dict(), sd_response):
            instance.fix_department("uuid", date(2020, 1, 1))

        call_list = instance.helper._mo_post.mock_calls

        first_mo_call = call(
            "details/edit",
            {
                "type": "org_unit",
                "data": {
                    "uuid": "uuid",
                    "user_key": instance.SD_DEPARTMENT_SHORTNAME,
                    "name": instance.SD_DEPARTMENT_NAME,
                    "parent": {"uuid": instance.MO_ORG_ROOT},
                    "org_unit_level": {"uuid": instance.MO_CLASS_UUID},
                    "org_unit_type": {"uuid": instance.MO_CLASS_UUID},
                    "validity": {"from": "2019-01-01", "to": "2023-12-31"},
                },
            },
        )
        second_mo_call = call(
            "details/edit",
            {
                "type": "org_unit",
                "data": {
                    "uuid": "uuid",
                    "user_key": instance.SD_DEPARTMENT_SHORTNAME,
                    "name": "new name",
                    "parent": {"uuid": instance.MO_ORG_ROOT},
                    "org_unit_level": {"uuid": instance.MO_CLASS_UUID},
                    "org_unit_type": {"uuid": instance.MO_CLASS_UUID},
                    "validity": {"from": "2024-01-01", "to": None},
                },
            },
        )

        assert first_mo_call in call_list
        assert second_mo_call in call_list

    def test_create_root_org_unit_if_unit_does_not_exists_in_mo(self):
        # Arrange
        instance = _TestableFixDepartments.get_instance()
        instance.get_parent = MagicMock(
            return_value=None
        )  # Meaning the unit is a root unit
        instance.helper.read_ou = MagicMock(return_value={"status": 404})
        instance._update_org_unit_for_single_sd_dep_registration = MagicMock()

        # Act
        with mock_sd_lookup("GetDepartment20111201", dict(), dict()):
            instance.fix_department(
                "99999999-9999-9999-9999-999999999999", datetime.today().date()
            )

        # Assert
        instance.helper._mo_post.assert_called_once_with(
            "ou/create",
            {
                "uuid": "99999999-9999-9999-9999-999999999999",
                "user_key": _TestableFixDepartments.SD_DEPARTMENT_SHORTNAME,
                "name": _TestableFixDepartments.SD_DEPARTMENT_NAME,
                "parent": {"uuid": _TestableFixDepartments.MO_ORG_ROOT},
                "org_unit_type": {"uuid": _TestableFixDepartments.MO_CLASS_UUID},
                "org_unit_level": {"uuid": _TestableFixDepartments.MO_CLASS_UUID},
                "validity": {
                    "from": "2019-01-01",
                    "to": None,
                },
            },
        )
        instance._update_org_unit_for_single_sd_dep_registration.assert_not_called()

    def test_create_parent_org_unit_if_unit_does_not_exists_in_mo(self):
        # Arrange
        instance = _TestableFixDepartments.get_instance()

        unit_uuid = "11111111-1111-1111-1111-111111111111"
        parent_uuid = "22222222-2222-2222-2222-222222222222"
        today = datetime.today().date()

        instance.get_parent = MagicMock(side_effect=[parent_uuid, None])
        instance.get_department = MagicMock(
            side_effect=[
                [
                    {
                        "DepartmentLevelIdentifier": _TestableFixDepartments.MO_CLASS_USER_KEY,
                        "DepartmentIdentifier": _TestableFixDepartments.SD_DEPARTMENT_SHORTNAME,
                        "DepartmentUUIDIdentifier": unit_uuid,
                        "DepartmentName": _TestableFixDepartments.SD_DEPARTMENT_NAME,
                        "ActivationDate": "2019-01-01",
                        "DeactivationDate": "9999-12-31",
                    }
                ],
                [
                    {
                        "DepartmentLevelIdentifier": _TestableFixDepartments.MO_CLASS_USER_KEY,
                        "DepartmentIdentifier": "Parent shortname",
                        "DepartmentUUIDIdentifier": parent_uuid,
                        "DepartmentName": "Parent",
                        "ActivationDate": "2019-01-01",
                        "DeactivationDate": "9999-12-31",
                    }
                ],
            ]
        )

        instance.helper.read_ou = MagicMock(side_effect=[dict(), {"status": 404}])

        expected_calls = [
            call(
                "ou/create",
                {
                    "uuid": "22222222-2222-2222-2222-222222222222",
                    "user_key": "Parent shortname",
                    "name": "Parent",
                    "parent": {"uuid": "00000000-0000-0000-0000-000000000000"},
                    "org_unit_type": {"uuid": _TestableFixDepartments.MO_CLASS_UUID},
                    "org_unit_level": {"uuid": _TestableFixDepartments.MO_CLASS_UUID},
                    "validity": {"from": "2019-01-01", "to": None},
                },
            ),
            call(
                "details/edit",
                {
                    "type": "org_unit",
                    "data": {
                        "uuid": "11111111-1111-1111-1111-111111111111",
                        "user_key": "some department short name",
                        "name": "some department name",
                        "parent": {"uuid": "22222222-2222-2222-2222-222222222222"},
                        "org_unit_level": {
                            "uuid": _TestableFixDepartments.MO_CLASS_UUID
                        },
                        "org_unit_type": {
                            "uuid": _TestableFixDepartments.MO_CLASS_UUID
                        },
                        "validity": {"from": "2019-01-01", "to": None},
                    },
                },
            ),
        ]

        # Act
        instance.fix_department(unit_uuid, today)

        # Assert
        actual_calls = instance.helper._mo_post.call_args_list
        assert expected_calls == actual_calls

        get_parent_calls = instance.get_parent.call_args_list
        assert get_parent_calls == [
            call("11111111-1111-1111-1111-111111111111", today),
            call("22222222-2222-2222-2222-222222222222", today),
        ]

    def test_fix_department_called_recursively(self):
        # Arrange
        instance = _TestableFixDepartments.get_instance()

        unit_uuid = "11111111-1111-1111-1111-111111111111"
        parent_uuid = "22222222-2222-2222-2222-222222222222"

        instance.get_parent = MagicMock(side_effect=[parent_uuid, None])
        instance.get_department = MagicMock(
            side_effect=[
                [
                    {
                        "DepartmentLevelIdentifier": _TestableFixDepartments.MO_CLASS_USER_KEY,
                        "DepartmentIdentifier": _TestableFixDepartments.SD_DEPARTMENT_SHORTNAME,
                        "DepartmentUUIDIdentifier": unit_uuid,
                        "DepartmentName": _TestableFixDepartments.SD_DEPARTMENT_NAME,
                        "ActivationDate": "2019-01-01",
                        "DeactivationDate": "9999-12-31",
                    }
                ],
                [
                    {
                        "DepartmentLevelIdentifier": _TestableFixDepartments.MO_CLASS_USER_KEY,
                        "DepartmentIdentifier": "Parent shortname",
                        "DepartmentUUIDIdentifier": parent_uuid,
                        "DepartmentName": "Parent",
                        "ActivationDate": "2019-01-01",
                        "DeactivationDate": "9999-12-31",
                    }
                ],
            ]
        )

        instance._create_org_unit_if_missing_in_mo = MagicMock(return_value=False)
        instance._update_org_unit_for_single_sd_dep_registration = MagicMock()

        # Act
        instance.fix_department(str(unit_uuid), date(2020, 1, 1))

        # Assert
        create_org_unit_calls = instance._create_org_unit_if_missing_in_mo.mock_calls
        update_calls = (
            instance._update_org_unit_for_single_sd_dep_registration.mock_calls
        )

        assert create_org_unit_calls == [
            call(
                {
                    "DepartmentLevelIdentifier": "Enhed",
                    "DepartmentIdentifier": "some department short name",
                    "DepartmentUUIDIdentifier": "11111111-1111-1111-1111-111111111111",
                    "DepartmentName": "some department name",
                    "ActivationDate": "2019-01-01",
                    "DeactivationDate": "9999-12-31",
                },
                parent_uuid,
            ),
            call(
                {
                    "DepartmentLevelIdentifier": "Enhed",
                    "DepartmentIdentifier": "Parent shortname",
                    "DepartmentUUIDIdentifier": "22222222-2222-2222-2222-222222222222",
                    "DepartmentName": "Parent",
                    "ActivationDate": "2019-01-01",
                    "DeactivationDate": "9999-12-31",
                },
                None,
            ),
        ]

        assert update_calls == [
            call(
                {
                    "DepartmentLevelIdentifier": "Enhed",
                    "DepartmentIdentifier": "Parent shortname",
                    "DepartmentUUIDIdentifier": parent_uuid,
                    "DepartmentName": "Parent",
                    "ActivationDate": "2019-01-01",
                    "DeactivationDate": "9999-12-31",
                },
                None,
            ),
            call(
                {
                    "DepartmentLevelIdentifier": "Enhed",
                    "DepartmentIdentifier": "some department short name",
                    "DepartmentUUIDIdentifier": unit_uuid,
                    "DepartmentName": "some department name",
                    "ActivationDate": "2019-01-01",
                    "DeactivationDate": "9999-12-31",
                },
                parent_uuid,
            ),
        ]

        get_parent_calls = instance.get_parent.call_args_list
        assert get_parent_calls == [
            call("11111111-1111-1111-1111-111111111111", date(2020, 1, 1)),
            call("22222222-2222-2222-2222-222222222222", date(2020, 1, 1)),
        ]

    def test_fix_ny_logic_elevates_engagement_from_too_deep_levels(self) -> None:
        """
        Test that an engagement is elevated from the "too deep" OU levels to
        the appropriate NY-levels above (happens when the SDTool button is
        pressed). Note: this was the first test of the "apply_NY_logic" function -
        we just test that (parts of it) is works "as is". We really should
        1) Find out what the function should actually do
        2) Test that the engagements dates are set correctly!
        """

        # Arrange
        unit_uuid = str(uuid4())
        parent_unit_uuid = str(uuid4())
        eng_uuid = str(uuid4())
        validity_date = date(2023, 8, 1)
        cpr = "0101901111"

        instance = _TestableFixDepartments.get_instance(
            {"sd_import_too_deep": ["Afdelings-niveau"]}
        )

        instance.helper.read_ou = MagicMock(
            return_value={
                "uuid": unit_uuid,
                "org_unit_level": {
                    "user_key": "Afdelings-niveau",
                },
                "parent": {
                    "uuid": parent_unit_uuid,
                    "org_unit_level": {
                        "user_key": "NY1-niveau",
                    },
                },
            }
        )
        instance._read_department_engagements = MagicMock(
            return_value={
                cpr: OrderedDict(
                    {
                        "PersonCivilRegistrationIdentifier": cpr,
                        "Employment": {
                            "EmploymentIdentifier": "12345",
                            "EmploymentDate": "2020-11-10",
                            "AnniversaryDate": "2004-08-15",
                            "EmploymentDepartment": {
                                "ActivationDate": "2020-11-10",
                                "DeactivationDate": "9999-12-31",
                                "DepartmentIdentifier": "department_id",
                                "DepartmentUUIDIdentifier": unit_uuid,
                            },
                            "Profession": {
                                "ActivationDate": "2020-11-10",
                                "DeactivationDate": "9999-12-31",
                                "JobPositionIdentifier": "2",
                                "EmploymentName": "Title",
                                "AppointmentCode": "0",
                            },
                            "EmploymentStatus": {
                                "ActivationDate": "2020-11-10",
                                "DeactivationDate": "9999-12-31",
                                "EmploymentStatusCode": "1",
                            },
                        },
                    }
                )
            }
        )
        instance.helper.read_user = MagicMock(
            return_value={
                "uuid": str(uuid4()),
            }
        )
        instance.helper.read_user_engagement = MagicMock(
            return_value=[
                {
                    "uuid": eng_uuid,
                    "org_unit": {"uuid": str(uuid4())},
                    "user_key": "12345",
                    "validity": {
                        "from": "2023-01-01",
                        "to": None,
                    },
                }
            ]
        )

        r = Response()
        r.status_code = 200

        instance.helper._mo_post.return_value = r

        # Act
        instance.fix_NY_logic(unit_uuid, validity_date)

        # Assert
        instance.helper._mo_post.assert_called_once_with(
            "details/edit",
            {
                "type": "engagement",
                "uuid": eng_uuid,
                "data": {
                    "org_unit": {"uuid": parent_unit_uuid},
                    "validity": {
                        "from": validity_date.strftime("%Y-%m-%d"),
                        "to": None,
                    },
                },
            },
        )



from datetime import date
from itertools import chain
from typing import Any
from typing import Callable
from typing import Dict
from typing import List
from typing import Tuple
from unittest import TestCase
from unittest.mock import MagicMock

from parameterized import parameterized

from ..ad_sync import AdMoSync
from ..utils import AttrDict
from .mocks import MO_USER_CPR
from .mocks import MO_UUID
from .mocks import MockADParameterReader
from .mocks import MockLoraCache
from .mocks import MockLoraCacheEmptyEmployee
from .mocks import MockMoraHelper
from .test_utils import dict_modifier
from .test_utils import TestADMoSyncMixin


def iso_date(date):
    return date.strftime("%Y-%m-%d")


def today_iso():
    return iso_date(date.today())


MO_VALUES = {"uuid": MO_UUID}


class _TestableAdMoSync(AdMoSync):
    def _setup_settings(self, all_settings):
        self.settings = {}

    def _setup_visibilities(self):
        self.visibility = {}

    def _setup_mora_helper(self):
        return MockMoraHelper(MO_USER_CPR)

    def _setup_lora_cache(self):
        return None


class _TestableAdMoSyncLoraCache(_TestableAdMoSync):
    def _setup_lora_cache(self):
        return MockLoraCache(MO_VALUES)


class _TestableAdMoSyncLoraCacheEmptyEmployee(_TestableAdMoSync):
    def _setup_lora_cache(self):
        return MockLoraCacheEmptyEmployee(MO_VALUES)


class _TestableAdMoSyncLoraCacheUserAttrs(_TestableAdMoSync):
    """A subclass of `_TestableAdMoSync` which is used by
    `TestEditUserAttrsLoraCache.test_edit_user_attrs_is_idempotent` to test that calls
    to `AdMoSync._edit_user_attrs` are idempotent.
    """

    class _MockMoraHelperMutatingLoraCache(MockMoraHelper):
        """A subclass of `MockMoraHelper` which records all calls to `update_user` and
        mutates the mocked (LoraCache) user object in order to test idempotency.
        """

        def __init__(self, loracache: MockLoraCache):
            self.update_user_calls: List[Tuple[str, Dict[str, Any]]] = []
            self._loracache = loracache
            super().__init__(MO_USER_CPR)

        def update_user(self, uuid, data):
            # Record attempt to mutate user given by `uuid`
            self.update_user_calls.append((uuid, data))
            # Mutate the user in mock `LoraCache`
            self._loracache.users[uuid][0].update(**data)

    def __init__(self):
        self.mo_user = {"uuid": MO_UUID}
        super().__init__()

    def _setup_settings(self, all_settings):
        # Configuration keys required by `AdMoSync._update_users`
        required_keys = {
            "ad_mo_sync_pre_filters": [],
            "ad_mo_sync_terminate_disabled": None,
            "ad_mo_sync_terminate_disabled_filters": [],
            "ad_mo_sync_terminate_missing": None,
            "ad_mo_sync_terminate_missing_require_itsystem": None,
        }

        # Map AD field "UserPrincipalName" to MO field "user_key"
        self.settings = {
            "integrations.ad": [
                {
                    "ad_mo_sync_mapping": {
                        "user_attrs": {"UserPrincipalName": "user_key"},
                    },
                    **required_keys,
                }
            ]
        }

    def _setup_lora_cache(self):
        return MockLoraCache(self.mo_user)

    def _setup_mora_helper(self):
        # Return a helper mutating the mock `LoraCache` returned by `_setup_lora_cache`
        return self._MockMoraHelperMutatingLoraCache(self.lc)

    def _setup_ad_reader_and_cache_all(self, index, cache_all=True):
        # Return a mock AD reader which also exposes our mocked `user_attrs` mapping

        def _get_setting():
            return self.settings["integrations.ad"][index]

        self.mock_ad_reader = MockADParameterReader()
        self.mock_ad_reader._get_setting = _get_setting
        return self.mock_ad_reader


class TestADMoSync(TestCase, TestADMoSyncMixin):
    maxDiff = None

    def setUp(self):
        self._initialize_configuration()

    def _sync_address_mapping_transformer(self):
        def add_sync_mapping(settings):
            settings["integrations.ad"][0]["ad_mo_sync_mapping"] = {
                "user_addresses": {
                    # Different visibility
                    "email": ["email_uuid", None],
                    "telephone": ["telephone_uuid", "PUBLIC"],
                    "office": ["office_uuid", "INTERNAL"],
                    "mobile": ["mobile_uuid", "SECRET"],
                    # No uuid
                    "floor": ["", None],
                }
            }
            return settings

        return add_sync_mapping

    def _setup_address_type_mappings(self, address_type):
        settings = self._prepare_settings(self._sync_address_mapping_transformer())
        mapping = settings["integrations.ad"][0]["ad_mo_sync_mapping"]
        address_type_setting = mapping["user_addresses"][address_type]
        return AttrDict(
            settings=settings,
            address_type_uuid=address_type_setting[0],
            address_type_visibility=address_type_setting[1],
        )

    def _get_expected_mo_api_calls(self, setup, expected, mo_values, ad_data, today):
        expected_sync = {
            "noop": [],
            "create": [
                {
                    "force": True,
                    "payload": {
                        "address_type": {"uuid": setup.address_type_uuid},
                        "org": {"uuid": "org_uuid"},
                        "person": {"uuid": mo_values["uuid"]},
                        "type": "address",
                        "validity": {"from": today, "to": None},
                        "value": ad_data,
                    },
                    "url": "details/create",
                }
            ],
            "edit": [
                {
                    "force": True,
                    "payload": [
                        {
                            "data": {
                                "address_type": {"uuid": setup.address_type_uuid},
                                "validity": {"from": today, "to": None},
                                "value": ad_data,
                            },
                            "type": "address",
                            "uuid": "address_uuid",
                        }
                    ],
                    "url": "details/edit",
                }
            ],
            "terminate": [
                {
                    "force": True,
                    "payload": {
                        "type": "address",
                        "uuid": "address_uuid",
                        "validity": {"to": today},
                    },
                    "url": "details/terminate",
                }
            ],
        }
        return expected_sync[expected]

    def _get_expected_mo_api_calls_with_visibility(
        self,
        setup: AttrDict,
        expected: str,
        mo_values: dict,
        ad_data: str,
        today: str,
    ):
        calls = self._get_expected_mo_api_calls(
            setup, expected, mo_values, ad_data, today
        )

        # Enrich expected with visibility
        address_type_visibility: str = setup.address_type_visibility  # type: ignore
        if address_type_visibility:
            # Where to write visibility information
            payload_table: Dict[str, Callable] = {
                "noop": lambda: {},  # aka. throw it away
                "create": lambda: calls[0]["payload"],
                "edit": lambda: payload_table["create"]()[0]["data"],
                "terminate": lambda: {},
            }
            # Write the visibility into the table
            visibility_lower = address_type_visibility.lower()
            payload_table[expected]()["visibility"] = {
                "uuid": "address_visibility_" + visibility_lower + "_uuid"
            }

        return calls

    def _get_expected_mo_engagement_edit_call(self, validity_from=None, **data):
        call = {
            "force": True,
            "payload": {
                "data": {
                    "validity": {"from": validity_from, "to": None},
                },
                "type": "engagement",
                "uuid": "engagement_uuid",
            },
            "url": "details/edit",
        }
        call["payload"]["data"].update(**data)
        return call

    @parameterized.expand(
        [
            # Email (Undefined)
            # ------------------
            # No email in MO
            ("email", None, None, "noop"),
            ("email", "emil@magenta.dk", None, "create"),
            ("email", "example@example.com", None, "create"),
            ("email", "lee@magenta.dk", None, "create"),
            # Email already in MO
            ("email", "emil@magenta.dk", "emil@magenta.dk", "noop"),
            ("email", "example@example.com", "emil@magenta.dk", "edit"),
            ("email", "lee@magenta.dk", "emil@magenta.dk", "edit"),
            # Email terminated in AD
            ("email", None, "old.mo.email@example.org", "terminate"),
            # Telephone number (PUBLIC)
            # --------------------------
            # No telephone number in MO
            ("telephone", None, None, "noop"),
            ("telephone", "+45 70 10 11 55", None, "create"),
            ("telephone", "70 10 11 55", None, "create"),
            ("telephone", "70101155", None, "create"),
            # Telephone number already in MO
            ("telephone", "70101155", "70101155", "noop"),
            ("telephone", "90909090", "70101155", "edit"),
            ("telephone", "90901111", "70101155", "edit"),
            # Telephone number terminated in AD
            ("telephone", None, "12345678", "terminate"),
            # Office number (INTERNAL)
            # -------------------------
            # No office number in MO
            ("office", None, None, "noop"),
            ("office", "420", None, "create"),
            ("office", "421", None, "create"),
            ("office", "11", None, "create"),
            # Office number already in MO
            ("office", "420", "420", "noop"),
            ("office", "421", "420", "edit"),
            ("office", "11", "420", "edit"),
            # Office number terminated in AD
            ("office", None, "42", "terminate"),
            # Mobile number (SECRET)
            # -----------------------
            # No mobile number in MO
            ("mobile", None, None, "noop"),
            ("mobile", "+45 70 10 11 55", None, "create"),
            ("mobile", "70 10 11 55", None, "create"),
            ("mobile", "70101155", None, "create"),
            # Mobile number already in MO
            ("mobile", "70101155", "70101155", "noop"),
            ("mobile", "90909090", "70101155", "edit"),
            ("mobile", "90901111", "70101155", "edit"),
            # Mobile number terminated in AD
            ("mobile", None, "12345678", "terminate"),
            # Floor (no uuid)
            # ----------------
            # No floor number in MO
            ("floor", None, None, "noop"),
            ("floor", "1st", None, "create"),
            ("floor", "2nd", None, "create"),
            ("floor", "3rd", None, "create"),
            # Floor number already in MO
            ("floor", "1st", "1st", "noop"),
            ("floor", "2nd", "1st", "edit"),
            ("floor", "3rd", "1st", "edit"),
            # Floor number terminated in AD
            ("floor", None, "1st", "terminate"),
        ]
    )
    def test_sync_address_data(self, address_type, ad_data, mo_data, expected):
        """Verify address data is synced correctly from AD to MO.

        Args:
            address_type (str): The type of address to operate on.
            ad_data (str): The address data found in AD (if any).
            mo_data (str): The address data found in MO (if any).
            expected (str): The expected outcome of running AD sync.
                One of:
                    'noop': Nothing in MO is updated.
                    'create': A new address is created in MO.
                    'edit': The current address in MO is updated.
                    'terminate': The current address in MO is terminated.
        """
        setup = self._setup_address_type_mappings(address_type)
        today = today_iso()
        mo_values = self.mo_values_func()

        # Helper functions to seed admosync mock
        def add_ad_data(ad_values):
            ad_values[address_type] = ad_data
            return ad_values

        def seed_mo():
            if mo_data is None:
                return {"address": []}
            return {
                "address": [
                    {
                        "uuid": "address_uuid",
                        "address_type": {"uuid": setup.address_type_uuid},
                        "org": {"uuid": "org_uuid"},
                        "person": {"uuid": mo_values["uuid"]},
                        "type": "address",
                        "validity": {"from": today, "to": None},
                        "value": mo_data,
                    }
                ]
            }

        self._setup_admosync(
            transform_settings=lambda _: setup.settings,
            transform_ad_values=add_ad_data,
            seed_mo=seed_mo,
        )

        self.assertEqual(self.ad_sync.mo_post_calls, [])

        # Run full sync against the mocks
        self.ad_sync.update_all_users()

        # Check that the expected MO calls were made
        self.assertEqual(
            self.ad_sync.mo_post_calls,
            self._get_expected_mo_api_calls_with_visibility(
                setup, expected, mo_values, ad_data, today
            ),
        )

    def test_sync_address_data_multiple_of_same_type(self):
        address_type = "email"
        setup = self._setup_address_type_mappings(address_type)
        today = today_iso()
        mo_values = self.mo_values_func()

        mo_value = "old value"
        ad_value = "new value"

        # Helper functions to seed admosync mock
        def add_ad_data(ad_values):
            ad_values[address_type] = ad_value
            return ad_values

        def seed_mo():
            three_identical_mo_addresses = [
                {
                    "uuid": "address_uuid",
                    "address_type": {"uuid": setup.address_type_uuid},
                    "org": {"uuid": "org_uuid"},
                    "person": {"uuid": mo_values["uuid"]},
                    "type": "address",
                    "validity": {"from": today, "to": None},
                    "value": mo_value,
                }
            ] * 3
            return {"address": three_identical_mo_addresses}

        self._setup_admosync(
            transform_settings=lambda _: setup.settings,
            transform_ad_values=add_ad_data,
            seed_mo=seed_mo,
        )

        self.assertEqual(self.ad_sync.mo_post_calls, [])

        # Run full sync against the mocks
        self.ad_sync.update_all_users()

        # Check that the expected MO calls were made.
        # Assert that we make 1 'edit' call and 2 'terminate' calls, in order
        # to keep only one of the three identical MO addresses.
        expected_calls = chain(
            *[
                self._get_expected_mo_api_calls(setup, verb, mo_values, ad_value, today)
                for verb in ("edit", "terminate", "terminate")
            ]
        )
        self.assertEqual(self.ad_sync.mo_post_calls, list(expected_calls))

    def test_sync_address_data_multiple(self):
        """Verify address data is synced correctly from AD to MO."""
        today = today_iso()
        mo_values = self.mo_values_func()

        # Helper functions to seed admosync mock
        def add_ad_data(ad_values):
            ad_values["email"] = "emil@magenta.dk"
            ad_values["telephone"] = "70101155"
            ad_values["office"] = "11"
            return ad_values

        def seed_mo():
            return {
                "address": [
                    {
                        "uuid": "address_uuid",
                        "address_type": {"uuid": "office_uuid"},
                        "org": {"uuid": "org_uuid"},
                        "person": {"uuid": mo_values["uuid"]},
                        "type": "address",
                        "validity": {"from": today, "to": None},
                        "value": "42",
                    }
                ]
            }

        self._setup_admosync(
            transform_settings=self._sync_address_mapping_transformer(),
            transform_ad_values=add_ad_data,
            seed_mo=seed_mo,
        )

        self.assertEqual(self.ad_sync.mo_post_calls, [])

        # Run full sync against the mocks
        self.ad_sync.update_all_users()

        # Expected outcome
        expected_sync = [
            {
                "force": True,
                "payload": {
                    "address_type": {"uuid": "email_uuid"},
                    "org": {"uuid": "org_uuid"},
                    "person": {"uuid": mo_values["uuid"]},
                    "type": "address",
                    "validity": {"from": today, "to": None},
                    "value": "emil@magenta.dk",
                },
                "url": "details/create",
            },
            {
                "force": True,
                "payload": {
                    "address_type": {"uuid": "telephone_uuid"},
                    "org": {"uuid": "org_uuid"},
                    "person": {"uuid": mo_values["uuid"]},
                    "type": "address",
                    "validity": {"from": today, "to": None},
                    "value": "70101155",
                    "visibility": {"uuid": "address_visibility_public_uuid"},
                },
                "url": "details/create",
            },
            {
                "force": True,
                "payload": [
                    {
                        "data": {
                            "address_type": {"uuid": "office_uuid"},
                            "validity": {"from": today, "to": None},
                            "value": "11",
                            "visibility": {"uuid": "address_visibility_internal_uuid"},
                        },
                        "type": "address",
                        "uuid": "address_uuid",
                    }
                ],
                "url": "details/edit",
            },
        ]
        self.assertEqual(len(self.ad_sync.mo_post_calls), 3)
        for expected_block in expected_sync:
            self.assertIn(expected_block, self.ad_sync.mo_post_calls)

    def _sync_itsystem_mapping_transformer(self):
        def add_sync_mapping(settings):
            settings["integrations.ad"][0]["ad_mo_sync_mapping"] = {
                "it_systems": {"samAccountName": "it_system_uuid"}
            }
            return settings

        return add_sync_mapping

    @parameterized.expand(
        [
            ("", "djjohn", "create"),
            ("", "dijon", "create"),
            ("djjohn", "djjohn", "noop"),
            ("djjohn", "dijon", "edit"),
            ("anything_else", "djjohn", "edit"),
            ("anything_else", "dijon", "edit"),
        ]
    )
    def test_sync_itsystem(self, mo_username, ad_username, expected):
        """Verify itsystem data is synced correctly from AD to MO."""
        today = today_iso()
        mo_values = self.mo_values_func()

        def seed_mo():
            from_date = "1970-01-01"
            to_date = "9999-12-31"
            return {
                "it": [
                    {
                        "itsystem": {
                            "name": "Active Directory",
                            "uuid": "it_system_uuid",
                        },
                        "user_key": mo_username,
                        "uuid": "it_system_uuid",
                        "validity": {"from": from_date, "to": to_date},
                    }
                ]
                if mo_username != ""
                else []
            }

        def set_ad_name(ad_values):
            ad_values["SamAccountName"] = ad_username
            return ad_values

        self._setup_admosync(
            transform_settings=self._sync_itsystem_mapping_transformer(),
            transform_ad_values=set_ad_name,
            seed_mo=seed_mo,
        )

        self.assertEqual(self.ad_sync.mo_post_calls, [])

        # Run full sync against the mocks
        self.ad_sync.update_all_users()

        # Expected outcome
        expected_sync = {
            "noop": [],
            "edit": [
                {
                    "force": True,
                    "payload": {
                        "type": "it",
                        "data": {
                            "user_key": ad_username,
                            "validity": {"from": today, "to": None},
                        },
                        "uuid": "it_system_uuid",
                    },
                    "url": "details/edit",
                }
            ],
            "create": [
                {
                    "force": True,
                    "payload": {
                        "type": "it",
                        "user_key": ad_username,
                        "itsystem": {"uuid": "it_system_uuid"},
                        "person": {"uuid": mo_values["uuid"]},
                        "validity": {"from": today, "to": None},
                    },
                    "url": "details/create",
                }
            ],
        }
        self.assertEqual(self.ad_sync.mo_post_calls, expected_sync[expected])

    def _sync_engagement_mapping_transformer(self, mo_to_ad):
        def add_sync_mapping(settings):
            settings["integrations.ad"][0]["ad_mo_sync_mapping"] = {
                "engagements": mo_to_ad
            }
            return settings

        return add_sync_mapping

    @parameterized.expand(
        [
            # Move GivenName into extensions_1
            (None, ("GivenName", "extension_1"), "noop"),
            ({"extension_1": None}, ("GivenName", "extension_1"), "edit"),
            ({"extension_1": "OldValue"}, ("GivenName", "extension_1"), "edit"),
            # Move GivenName into extensions_2
            (None, ("GivenName", "extension_2"), "noop"),
            ({"extension_2": None}, ("GivenName", "extension_2"), "edit"),
            ({"extension_2": "OldValue"}, ("GivenName", "extension_2"), "edit"),
            # Move SamAccountName into extensions_1
            (None, ("SamAccountName", "extension_1"), "noop"),
            ({"extension_1": None}, ("SamAccountName", "extension_1"), "edit"),
            ({"extension_1": "OldValue"}, ("SamAccountName", "extension_1"), "edit"),
        ]
    )
    def test_sync_engagement(self, do_seed, mo_to_ad, expected):
        """Verify engagement data is synced correctly from AD to MO."""
        today = today_iso()
        ad_values = self.ad_values_func()

        def seed_mo():
            if not do_seed:
                return {"engagement": []}
            element = {
                "is_primary": True,
                "uuid": "engagement_uuid",
                "validity": {"from": "1960-06-29", "to": None},
            }
            element.update(do_seed)
            return {"engagement": [element]}

        self._setup_admosync(
            transform_settings=self._sync_engagement_mapping_transformer(
                {key: value for (key, value) in [mo_to_ad]}
            ),
            seed_mo=seed_mo,
        )

        self.assertEqual(self.ad_sync.mo_post_calls, [])

        # Run full sync against the mocks
        self.ad_sync.update_all_users()

        # Expected outcome
        expected_sync = {
            "noop": [],
            "edit": [
                {
                    "force": True,
                    "payload": {
                        "data": {
                            mo_to_ad[1]: ad_values[mo_to_ad[0]],
                            "validity": {"from": today, "to": None},
                        },
                        "type": "engagement",
                        "uuid": "engagement_uuid",
                    },
                    "url": "details/edit",
                }
            ],
        }
        self.assertEqual(self.ad_sync.mo_post_calls, expected_sync[expected])

    def test_sync_engagement_dropped_field(self):
        """Verify engagement data is synced correctly from AD to MO.
        Test removal of dropped extensions in incoming AD object.

        This tests the code path in `_edit_engagement` where `self.lc` is
        None, e.g. the LoraCache is not configured and used.
        """
        today = today_iso()

        def seed_mo():
            element = {
                "is_primary": True,
                "uuid": "engagement_uuid",
                "validity": {"from": "1960-06-29", "to": None},
                "extension_2": "old mo value",
            }
            return {"engagement": [element]}

        self._setup_admosync(
            transform_settings=self._sync_engagement_mapping_transformer(
                # Map an attr which does not exist in the AD object
                {"extensionAttribute2": "extension_2"}
            ),
            seed_mo=seed_mo,
        )

        self.assertEqual(self.ad_sync.mo_post_calls, [])

        # Run full sync against the mocks
        self.ad_sync.update_all_users()

        # Verify that we send an empty value to MO for 'extension_2'
        self.assertEqual(
            self.ad_sync.mo_post_calls,
            [
                self._get_expected_mo_engagement_edit_call(
                    extension_2=None, validity_from=today
                )
            ],
        )

    def test_sync_engagement_dropped_field_loracache(self):
        """Verify engagement data is synced correctly from AD to MO.
        Test removal of dropped extensions in incoming AD object.

        This test mocks the presence of a `LoraCache` instance at `self.lc`.
        """
        self._setup_admosync(
            transform_settings=self._sync_engagement_mapping_transformer(
                # Map an attr which does not exist in the AD object
                {"extensionAttribute2": "extension_2"}
            ),
        )

        self.assertEqual(self.ad_sync.mo_post_calls, [])

        self.ad_sync.lc = MockLoraCache(self.mo_values_func())

        # Run full sync against the mocks
        self.ad_sync.update_all_users()

        # Verify that we send an empty value to MO for 'extension_2'
        self.assertEqual(
            self.ad_sync.mo_post_calls,
            [
                self._get_expected_mo_engagement_edit_call(
                    extension_2=None,
                    validity_from=today_iso(),
                )
            ],
        )

    def test_sync_engagement_configuration_check_loracache(self):
        """Verify that `_edit_engagement` raises an exception if an
        unmapped MO field is encountered.

        This test mocks the presence of a `LoraCache` instance at `self.lc`.
        """
        self._setup_admosync(
            transform_settings=self._sync_engagement_mapping_transformer(
                # Map an attr which does not exist in field mapping
                {"extensionAttribute2": "unknown_field"}
            ),
        )

        self.ad_sync.lc = MockLoraCache(self.mo_values_func())

        # Run full sync against the mocks
        with self.assertRaises(Exception):
            self.ad_sync.update_all_users()

    @parameterized.expand(
        [
            (False,),  # test without LoraCache
            (True,),  # test with LoraCache
        ]
    )
    def test_sync_engagement_excludes_future_engagements(self, lora_cache):
        def seed_mo():
            current = {
                "uuid": "engagement_uuid_1",
                "extension_1": "1",
                "is_primary": True,
                "validity": {"from": "2010-12-31", "to": "2030-12-31"},
            }
            future = {
                "uuid": "engagement_uuid_2",
                "extension_1": "2",
                "is_primary": True,
                "validity": {"from": "2030-12-31", "to": None},
            }
            return {"engagement": [current, future]}

        self._setup_admosync(
            transform_settings=self._sync_engagement_mapping_transformer(
                # Map an attr whose value is not equal to the MO value
                {"ad_field": "extension_1"}
            ),
            # Mock an AD which always contains "foobar" in the field
            # "ad_field"
            transform_ad_values=dict_modifier({"ad_field": "foobar"}),
            seed_mo=seed_mo,
        )

        if lora_cache:
            self.ad_sync.lc = MockLoraCache(
                self.mo_values_func(),
                mo_engagements=seed_mo()["engagement"],
            )

        # Run full sync against the mocks
        self.ad_sync.update_all_users()

        self.assertEqual(
            self.ad_sync.mo_post_calls,
            [
                {
                    "force": True,
                    "payload": {
                        "data": {
                            "extension_1": "foobar",
                            "validity": {
                                "from": today_iso(),
                                "to": "2030-12-31",
                            },
                        },
                        "type": "engagement",
                        "uuid": "engagement_uuid_1",
                    },
                    "url": "details/edit",
                }
            ],
        )

    @parameterized.expand(
        [
            # - Finalize
            # Today
            [today_iso(), None, "terminate"],
            [today_iso(), today_iso(), "noop"],
            # 2020-01-01
            ["2020-01-01", None, "terminate"],
            ["2020-01-01", today_iso(), "noop"],
            ["2020-01-01", "2020-02-01", "noop"],  # past
            ["2020-01-01", "9999-01-01", "noop"],  # future
        ]
    )
    def test_finalization_address(self, from_date, to_date, expected):
        """Verify expected behavior from sync_disabled settings."""

        def add_sync_mapping(settings):
            settings["integrations.ad"][0]["ad_mo_sync_mapping"] = {
                "user_addresses": {"email": ["email_uuid", None]}
            }
            settings["integrations.ad"][0]["ad_mo_sync_terminate_disabled"] = True
            return settings

        mo_values = self.mo_values_func()

        # Helper functions to seed admosync mock
        def add_ad_data(ad_values):
            ad_values["Enabled"] = False
            return ad_values

        def seed_mo():
            return {
                "address": [
                    {
                        "uuid": "address_uuid",
                        "address_type": {"uuid": "email_uuid"},
                        "org": {"uuid": "org_uuid"},
                        "person": {"uuid": mo_values["uuid"]},
                        "type": "address",
                        "validity": {"from": from_date, "to": to_date},
                        "value": "42",
                    }
                ]
            }

        self._setup_admosync(
            transform_settings=add_sync_mapping,
            transform_ad_values=add_ad_data,
            seed_mo=seed_mo,
        )

        self.assertEqual(self.ad_sync.mo_post_calls, [])

        # Run full sync against the mocks
        self.ad_sync.update_all_users()

        today = today_iso()
        sync_expected = {
            "terminate": [
                {
                    "force": True,
                    "payload": {
                        "type": "address",
                        "uuid": "address_uuid",
                        "validity": {"to": today},
                    },
                    "url": "details/terminate",
                }
            ],
            "noop": [],
        }
        self.assertEqual(self.ad_sync.mo_post_calls, sync_expected[expected])

    @parameterized.expand(
        [
            # - Finalize
            # Today
            [today_iso(), None, "terminate"],
            [today_iso(), today_iso(), "noop"],
            # 2020-01-01
            ["2020-01-01", None, "terminate"],
            ["2020-01-01", today_iso(), "noop"],
            ["2020-01-01", "2020-02-01", "noop"],  # past
            ["2020-01-01", "9999-01-01", "noop"],  # future
        ]
    )
    def test_finalization_itsystem(self, from_date, to_date, expected):
        """Verify expected behavior from sync_disabled settings."""

        def add_sync_mapping(settings):
            settings["integrations.ad"][0]["ad_mo_sync_mapping"] = {
                "it_systems": {"samAccountName": "it_system_uuid"}
            }
            settings["integrations.ad"][0]["ad_mo_sync_terminate_disabled"] = True
            return settings

        # Helper functions to seed admosync mock
        def add_ad_data(ad_values):
            ad_values["Enabled"] = False
            return ad_values

        def seed_mo():
            return {
                "it": [
                    {
                        "itsystem": {
                            "name": "Active Directory",
                            "uuid": "it_system_uuid",
                        },
                        "user_key": "username",
                        "uuid": "itconnection_uuid",
                        "validity": {"from": from_date, "to": to_date},
                    }
                ]
            }

        self._setup_admosync(
            transform_settings=add_sync_mapping,
            transform_ad_values=add_ad_data,
            seed_mo=seed_mo,
        )

        self.assertEqual(self.ad_sync.mo_post_calls, [])

        # Run full sync against the mocks
        self.ad_sync.update_all_users()

        today = today_iso()
        sync_expected = {
            "terminate": [
                {
                    "force": True,
                    "payload": {
                        "type": "it",
                        "uuid": "itconnection_uuid",
                        "validity": {"to": today},
                    },
                    "url": "details/terminate",
                }
            ],
            "noop": [],
        }
        self.assertEqual(self.ad_sync.mo_post_calls, sync_expected[expected])

    @parameterized.expand(
        [
            ([], True),
            (["True"], True),
            (["False"], False),
            (["True", "True"], True),
            (["True", "False"], False),
            (["False", "True"], False),
            (["False", "False"], False),
            (["{{ 'a' == 'a' }}"], True),
            (["{{ 'a' == 'b' }}"], False),
            (["{{ 2 == 3 }}"], False),
            (["{{ ad_object['extensionAttribute1']|length == 10 }}"], True),
        ]
    )
    def test_pre_filters(self, prefilters, expected):
        def add_prefilter_template(settings):
            settings["integrations.ad"][0]["ad_mo_sync_mapping"] = {}
            settings["integrations.ad"][0]["ad_mo_sync_pre_filters"] = prefilters
            return settings

        self._setup_admosync(
            transform_settings=add_prefilter_template,
        )
        self.assertEqual(self.ad_sync.mo_post_calls, [])

        update_mock = MagicMock()
        self.ad_sync._update_single_user = update_mock

        # Run full sync against the mocks
        self.ad_sync.update_all_users()

        if expected:
            update_mock.assert_called()
        else:
            update_mock.assert_not_called()

    @parameterized.expand(
        [
            ([], False, False),
            (["True"], False, False),
            (["False"], False, False),
            ([], True, True),
            (["True"], True, True),
            (["False"], True, True),
            ([], None, False),
            (["True"], None, True),
            (["False"], None, False),
            (["True", "True"], None, True),
            (["True", "False"], None, False),
            (["False", "True"], None, False),
            (["False", "False"], None, False),
        ]
    )
    def test_disabled_filter(self, prefilters, terminate_disabled, expected):
        def add_terminate_filter_template(settings):
            settings["integrations.ad"][0]["ad_mo_sync_mapping"] = {}
            settings["integrations.ad"][0][
                "ad_mo_sync_terminate_disabled"
            ] = terminate_disabled
            settings["integrations.ad"][0][
                "ad_mo_sync_terminate_disabled_filters"
            ] = prefilters
            return settings

        # Helper functions to seed admosync mock
        def add_ad_data(ad_values):
            ad_values["Enabled"] = False
            return ad_values

        self._setup_admosync(
            transform_settings=add_terminate_filter_template,
            transform_ad_values=add_ad_data,
        )
        self.assertEqual(self.ad_sync.mo_post_calls, [])

        finalize_mock = MagicMock()
        self.ad_sync._finalize_it_system = finalize_mock
        self.ad_sync._finalize_user_addresses = finalize_mock

        # Run full sync against the mocks
        self.ad_sync.update_all_users()

        if expected:
            finalize_mock.assert_called()
        else:
            finalize_mock.assert_not_called()

    @parameterized.expand(
        [
            (False, False),
            (True, True),
        ]
    )
    def test_finalize_missing_ad_user(self, terminate_missing, expected):
        def add_terminate_filter_template(settings):
            settings["integrations.ad"][0]["ad_mo_sync_mapping"] = {}
            settings["integrations.ad"][0][
                "ad_mo_sync_terminate_missing"
            ] = terminate_missing
            settings["integrations.ad"][0][
                "ad_mo_sync_terminate_missing_require_itsystem"
            ] = False
            return settings

        # Helper functions to seed admosync mock
        def add_ad_data(ad_values):
            return None

        self._setup_admosync(
            transform_settings=add_terminate_filter_template,
            transform_ad_values=add_ad_data,
        )
        self.assertEqual(self.ad_sync.mo_post_calls, [])

        finalize_mock = MagicMock()
        engagement_mock = MagicMock()
        self.ad_sync._finalize_it_system = finalize_mock
        self.ad_sync._finalize_user_addresses = finalize_mock
        self.ad_sync._edit_engagement = engagement_mock

        # Run full sync against the mocks
        self.ad_sync.update_all_users()

        if expected:
            finalize_mock.assert_called()
            engagement_mock.assert_called()
        else:
            finalize_mock.assert_not_called()


class TestADMoSyncEditUserAttrs(TestCase, TestADMoSyncMixin):
    maxDiff = None
    generate_dynamic_person = False

    def setUp(self):
        self._initialize_configuration()

    @parameterized.expand(
        [
            # 0: No fields are mapped
            (
                {},  # (empty) mapping of AD attrs to MO attrs
                {},  # (additional) AD attrs for this user
            ),
            # 1: "Given name" is mapped from AD to MO, and AD value is present
            (
                {"givenName": "givenname"},
                {"givenName": "Test"},
            ),
            # 2: "Given name" is mapped from AD to MO, but AD value is not
            # present
            (
                {"givenName": "givenname"},
                {},
            ),
            # 3: Field A is mapped from AD to MO, but only field B has a
            # present AD value
            (
                {"foo": "foo"},
                {"bar": "AD value for bar"},
            ),
            # 4: Fields A and B are mapped, but only field B has a present AD
            # value
            (
                {"givenName": "givenname", "surname": "surname"},
                {"givenName": "Test"},
            ),
            # 5: Fields A and B are mapped, and both fields have a present AD
            # value
            (
                {"givenName": "givenname", "surname": "surname"},
                {"givenName": "Test", "surname": "Testesen"},
            ),
        ]
    )
    def test_edit_user_attrs(self, mapping, ad_attrs):
        def add_sync_mapping(settings):
            settings["integrations.ad"][0]["ad_mo_sync_mapping"] = {
                "user_attrs": mapping
            }
            return settings

        def add_ad_data(ad_values):
            ad_values.update(ad_attrs)
            return ad_values

        self._setup_admosync(
            transform_settings=add_sync_mapping,
            transform_ad_values=add_ad_data,
        )

        # Run full sync against the mocks
        self.ad_sync.update_all_users()

        # If AD attrs are mapped and present in AD object
        if set(ad_attrs) and set(ad_attrs).issubset(set(mapping)):
            # Assert exactly one MO POST call was made
            self.assertEqual(len(self.ad_sync.mo_post_calls), 1)
            call = self.ad_sync.mo_post_calls[0]
            data = call["payload"]["data"]
            # Assert payload is "edit employee" with validity of (today, None)
            self.assertEqual(call["url"], "details/edit")
            self.assertEqual(call["payload"]["type"], "employee")
            self.assertDictEqual(data["validity"], {"from": today_iso(), "to": None})
            # Assert we use AD value for each mapped field
            for ad_attr, mo_attr in mapping.items():
                if ad_attr in ad_attrs:
                    self.assertEqual(data[mo_attr], ad_attrs[ad_attr])
                else:
                    self.assertNotIn(ad_attr, data)
        else:
            # Assert no MO POST calls were made
            self.assertEqual(len(self.ad_sync.mo_post_calls), 0)


class TestReadAllMOUsers(TestCase):
    def test_returns_users(self):
        """`AdMoSync._read_all_mo_users` must return all non-empty users found in
        `self.lc.users` if using LoraCache.

        See: #48568
        """
        instance = _TestableAdMoSyncLoraCache()
        expected_users = [MO_VALUES]
        actual_users = instance._read_all_mo_users()
        self.assertEqual(actual_users, expected_users)

    def test_handles_empty_loracache_user(self):
        """Regression test. `AdMoSync._read_all_mo_users` did not handle "empty
        LoraCache users" well, and would fail with an `IndexError`.

        (An "empty LoraCache user" is a key/value pair where the key is a valid UUID but
        the value is an empty list.)

        See: #48568, #50410
        """
        instance = _TestableAdMoSyncLoraCacheEmptyEmployee()
        employees = instance._read_all_mo_users()
        self.assertEqual(employees, [])


class TestEditUserAttrsLoraCache:
    def test_edit_user_attrs_is_idempotent(self):
        # Invoke `_edit_user_attrs` twice on the same MO user and AD user
        instance = _TestableAdMoSyncLoraCacheUserAttrs()
        instance._update_users([instance.mo_user])
        instance._update_users([instance.mo_user])

        # Assert that we only issue *one* call to `self.helper.update_user` in
        # `_edit_user_attrs` as neither the AD user or MO user has changed between the
        # two update attempts.
        assert len(instance.helper.update_user_calls) == 1

        # Assert contents of the single update
        mo_uuid, mo_data = instance.helper.update_user_calls[0]
        ad_user = instance.mock_ad_reader.read_user()
        assert mo_uuid == MO_UUID
        assert mo_data["user_key"] == ad_user["UserPrincipalName"]
        assert mo_data["validity"] == {"from": today_iso(), "to": None}



import logging
from uuid import UUID

import click
from pydantic import ValidationError

from .config import get_changed_at_settings
from .sd_common import sd_lookup


class TestSdConnectivity(object):
    def __init__(self):
        self.validation_error = None
        try:
            self.settings = get_changed_at_settings()
            self.settings.job_settings.start_logging_based_on_settings()
        except ValidationError as err:
            self.validation_error = err

    def _check_sd_settings(self):
        logging.info("Check settings...")

        if self.validation_error is None:
            logging.info("SD settings OK")
        else:
            logging.warning(
                "The following SD settings errors were detected: %s",
                self.validation_error,
            )

        logging.info("Done checking SD settings")

    def _check_contact_to_sd(self):
        logging.info("Tjekker at vi har kontakt til SD")
        params = {
            "UUIDIndicator": "true",
            "InstitutionIdentifier": self.settings.sd_institution_identifier,
        }
        try:
            institution_info = sd_lookup(
                "GetInstitution20111201",
                settings=self.settings,
                params=params,
            )
        except Exception:
            logging.exception("Fejl i kontakt til SD Løn")
            exit(1)

        try:
            institution = institution_info["Region"]["Institution"]
            institution_uuid = institution["InstitutionUUIDIdentifier"]
            UUID(institution_uuid, version=4)
            logging.info("Korrekt kontakt til SD Løn")
        except Exception:
            logging.exception(
                "Fik forbindelse, men modtog ikke-korrekt svar fra SD: %s",
                institution_info,
            )
            exit(1)

    def sd_check(self):
        self._check_sd_settings()
        self._check_contact_to_sd()


@click.command()
def check_connectivity():
    """Check SD configuration and connectivity."""
    tsc = TestSdConnectivity()
    tsc.sd_check()


if __name__ == "__main__":
    check_connectivity()



import logging
import random
import time
from operator import itemgetter

from ra_utils.tqdm_wrapper import tqdm

from .ad_common import AD

logger = logging.getLogger("AdReader")


# SKIP_BRUGERTYPE


def first_included(settings, users):
    """
    include: given a list of users, return the first one that is included
    exclude: given a list of users, return the first one that is not excluded
    """
    discrim_field = settings.get("discriminator.field")

    if not discrim_field:
        return next(iter(users), {})

    if settings["discriminator.function"] == "include":
        users = list(filter(lambda user: discrim_field in user, users))

    for value in settings.get("discriminator.values", []):
        for user in users:
            value_found = value == user.get(discrim_field)
            if settings["discriminator.function"] == "include" and value_found:
                return user
            if settings["discriminator.function"] == "exclude" and not value_found:
                return user
    return {}


class ADParameterReader(AD):
    def read_encoding(self):
        """
        Read the character encoding of the Power Shell session.
        """
        ps_script = "$OutputEncoding | ConvertTo-Json"
        response = self._run_ps_script(ps_script)
        return response

    def read_it_all(self, print_progress=False):
        return self.cache_all(print_progress=print_progress)

    # Hvornår skal vi læse og skrive i hvad?
    #
    # Opdatering af OS2mo fra AD:
    # -> ad_sync: vi læser alle (for i in...) Ny reader og opdater mo
    #
    # Opdatering af AD fra OS2mo:
    # -> mo_to_ad_sync: skriv kun til det første (og læs kun fra det første)
    #
    # Indlæsning fra Opus og SD:
    # -> Her læser vi fra 0'eren - vi skal have fat i object-guid
    # cpr-uid-map-til rollekatalog skal også kun læse fra 0 (for now....)
    # vi kan teste en del med den, altså:
    # -> venv/bin/python exporters/cpr_uuid.py --use-ad
    #
    # cpr_mo_ad_map.csv har kun uuider/brugernavne pǻ de linier, hvor den
    # sd-importerede bruger også er i AD.

    def uncached_read_user(self, user=None, cpr=None, ria=None):
        # read one or more users using cpr-pattern.
        # if list is passed in ria (read it all) then this is extended
        # with found users - this way the function replaces the old
        # 'read it all' function, so there is now only one function
        # reading from AD.
        settings = self._get_setting()
        cpr_field = settings["cpr_field"]
        cpr_separator = settings.get("cpr_separator", "")
        caseless_samname = settings.get("caseless_samname", False)
        sam_filter = settings.get("sam_filter", "")

        logger.debug(f"Uncached AD read, user {user}")

        server = None
        if self.all_settings["primary"]["servers"]:
            server = random.choice(self.all_settings["primary"]["servers"])

        response = self.get_from_ad(user=user, cpr=cpr, server=server)

        users_by_cpr = {}
        for user in response:
            users_by_cpr.setdefault(user[settings["cpr_field"]], []).append(user)

        try:
            for userlist in users_by_cpr.values():
                current_user = first_included(settings, userlist)
                if current_user:
                    current_user_samaccountname = current_user["SamAccountName"]
                    self.results[current_user_samaccountname] = current_user

                    # Remove CPR separator if found
                    cpr = current_user.get(cpr_field, "")
                    if cpr is not None:
                        cpr = cpr.replace(cpr_separator, "")

                    # Filter on SamAccountName
                    if caseless_samname:
                        if current_user_samaccountname.lower().startswith(
                            sam_filter.lower()
                        ):
                            self.results[cpr] = current_user
                    else:
                        if current_user_samaccountname.startswith(sam_filter):
                            self.results[cpr] = current_user

                    # Store in accumulator variable, if given
                    if ria is not None:
                        ria.append(current_user)

        except Exception:
            logger.error("Response from uncached_read_user: {}".format(response))
            raise

    def cache_all(self, print_progress=False):
        logger.info("Caching all users")
        t = time.time()
        return_value = []
        date_range = list(range(1, 32))
        if pseudo_cprs := self.all_settings["primary"].get("pseudo_cprs"):
            date_range.extend(pseudo_cprs)
        if print_progress:
            date_range = tqdm(date_range, desc="Fetching AD accounts")
        for i in date_range:
            day = str(i).zfill(2)
            self.uncached_read_user(cpr="{}*".format(day), ria=return_value)
            logger.debug(len(self.results))
            logger.debug("Read time: {}".format(time.time() - t))
        return return_value

    def read_user(self, user=None, cpr=None, cache_only=False):
        """Read all properties of an AD user.

        The user can be retrived either by cpr or by AD user name.

        :param user: The AD username to retrive.
        :param cpr: CPR number of the user to retrive.
        :param cache_only: Return {} if user is not already cached
        :return: All properties listed in AD for the user.
        """
        logger.debug(f"Cached AD read, user {user}")
        if (not cpr) and (not user):
            return

        if user:
            dict_key = user
            if user in self.results:
                return self.results[user]

        if cpr:
            dict_key = cpr
            if cpr in self.results:
                return self.results[cpr]

        if cache_only:
            return {}

        # Populate self.results:
        self.uncached_read_user(user=user, cpr=cpr)

        logger.debug(
            "Returned info for {}: {}".format(dict_key, self.results.get(dict_key, {}))
        )
        return self.results.get(dict_key, {})

    def get_all_samaccountname_values(self):
        """Retrieve *all* AD usernames (= "SamAccountName" values) including usernames
        of AD users without a CPR or MO user UUID.
        """
        server = self.all_settings["global"]["servers"][0]
        cmd = f"""{self._build_user_credential()}
        Get-ADUser -Credential $usercredential -Filter '*' -Server {server} -Properties SamAccountName | ConvertTo-Json"""
        all_users = self._run_ps_script(cmd)
        return set(map(itemgetter("SamAccountName"), all_users))

    def get_all_email_values(self):
        """Retrieve *all* AD email addresses, including emails of AD users without a CPR
        or MO user UUID.
        """
        server = self.all_settings["global"]["servers"][0]
        cmd = f"""{self._build_user_credential()}
        Get-ADUser -Credential $usercredential -Filter '*' -Server {server} -Properties EmailAddress | ConvertTo-Json"""
        all_users = self._run_ps_script(cmd)
        return set(map(itemgetter("EmailAddress"), all_users))



import logging
from typing import Any
from typing import OrderedDict

from more_itertools import one

from .config import get_changed_at_settings

logger = logging.getLogger("sdChangedAt")


def cpr_env_filter(entity: OrderedDict[str, Any]) -> bool:
    settings = get_changed_at_settings()

    cpr = entity["PersonCivilRegistrationIdentifier"]

    if settings.sd_exclude_cprs_mode:
        # The CPRs in the sd_cprs should be excluded from further processing
        # and all other CPRs should be processed
        process_cpr = cpr not in settings.sd_cprs
    else:
        # In this case "exclude mode" is False, i.e. "include mode" is True
        # which means that the cprs in sd_cprs are the ONLY ones that should
        # be processed further
        process_cpr = cpr in settings.sd_cprs

    if not process_cpr and settings.sd_exclude_cprs_mode:
        logger.warning(f"*** SKIPPING employee with cpr={cpr[:6]} ***")

    return process_cpr


def skip_fictional_users(entity) -> bool:
    cpr = entity["PersonCivilRegistrationIdentifier"]
    if cpr[-4:] == "0000":
        logger.warning("Skipping fictional user: {}".format(cpr))
        return False
    return True


def skip_job_position_id(
    profession: OrderedDict[str, Any], job_pos_ids_to_skip: list[str]
) -> bool:
    """
    Check if SD JobPositionIdentifier is in the list to skip,
    i.e. the list provided via the environment variable
    SD_SKIP_EMPLOYMENT_TYPES

    Args:
        profession: a "Profession" in the list of professions in the
          SD employment.
        job_pos_ids_to_skip: list of SD JobPositionIdentifiers to skip

    Returns:
        True if the SD profession should be skipped and false otherwise.
    """

    job_pos_id = profession.get("JobPositionIdentifier")
    if job_pos_id in job_pos_ids_to_skip:
        return True

    return False



import unittest
from copy import deepcopy
from pathlib import Path
from unittest import TestCase

from hypothesis import given
from hypothesis.strategies import text
from parameterized import parameterized

from integrations.opus import opus_helpers


testfile1 = Path.cwd() / "integrations/opus/tests/ZLPE20200101_delta.xml"
testfile2 = Path.cwd() / "integrations/opus/tests/ZLPE20200102_delta.xml"


class test_opus_helpers(TestCase):
    @given(text())
    def test_generate_uuid(self, value):
        uuid1 = opus_helpers.generate_uuid(value)
        uuid2 = opus_helpers.generate_uuid(value)
        self.assertEqual(uuid1, uuid2)

    @parameterized.expand(
        [
            (testfile1,),
            (testfile2,),
        ]
    )
    def test_file_diff_same(self, file):
        file_diffs = opus_helpers.file_diff(file, file, disable_tqdm=True)
        self.assertEqual(file_diffs["units"], [])
        self.assertEqual(file_diffs["employees"], [])

    def test_file_diff(self):
        file_diffs = opus_helpers.file_diff(testfile1, testfile2, disable_tqdm=True)
        self.assertNotEqual(file_diffs["units"], [])
        self.assertNotEqual(file_diffs["employees"], [])

    @parameterized.expand(
        [
            (testfile1,),
            (testfile2,),
        ]
    )
    def test_parser(self, file):
        self.units, self.employees = opus_helpers.parser(file)
        self.assertIsInstance(self.units, list)
        self.assertIsInstance(self.employees, list)
        self.assertIsInstance(self.units[0], dict)
        self.assertIsInstance(self.employees[0], dict)

    @parameterized.expand(
        [
            (testfile1, [], set()),
            (testfile2, [], set()),
            (testfile1, ["1"], {"1", "2", "3"}),
            (testfile2, ["1"], {"1", "2", "3", "4", "5"}),
        ]
    )
    def test_all_filtered_ids_fiter_top_unit(self, file, filter_ids, expected):
        all_filtered = opus_helpers.find_all_filtered_ids(file, filter_ids)
        self.assertEqual(len(all_filtered), len(expected))
        self.assertEqual(all_filtered, expected)

    @parameterized.expand(
        [
            (testfile1, 1),
            (testfile2, 2),
        ]
    )
    def test_split_employees_active(self, inputfile, expected):
        file_diffs = opus_helpers.file_diff(None, inputfile, disable_tqdm=True)
        employees, leaves = opus_helpers.split_employees_leaves(file_diffs["employees"])
        employees = list(employees)
        leaves = list(leaves)
        self.assertEqual(len(leaves), expected)

    @parameterized.expand(
        [
            ([], [3, 0, 4, 1], None),
            (["1"], [0, 3, 0, 1], None),
            # Read specific opus_id
            ([], [1, 0, 0, 0], 1),
            (["1"], [0, 1, 0, 0], 1),
        ]
    )
    def test_full_(self, filter_ids, expected, opus_id):
        data = opus_helpers.read_and_transform_data(
            None,
            testfile1,
            filter_ids,
            disable_tqdm=True,
            opus_id=opus_id,
        )
        # data is a tuple of units, filtered units, employess, terminated employees
        # test that the length of each is as expected
        for i, x in enumerate(data):
            this = list(x)
            assert (
                len(this) == expected[i]
            ), f"Expected {expected[i]} objects, got {len(this) }\n{this}"

    def test_find_changed_parent(self):
        org1, _ = opus_helpers.parser(testfile1)
        org2 = deepcopy(org1)
        org2[2]["parentOrgUnit"] = "CHANGED!"
        diffs = opus_helpers.find_changes(before=org1, after=org2, disable_tqdm=True)
        self.assertEqual(diffs, [org2[2]])

    def test_find_cancelled(self):
        file_diffs = opus_helpers.file_diff(testfile1, testfile2, disable_tqdm=True)
        assert (
            len(file_diffs["cancelled_employees"]) == 1
        ), "Expected to find 1 cancelled employee"


if __name__ == "__main__":
    unittest.main()



from datetime import date
from functools import cache
from uuid import UUID

from gql import gql
from more_itertools import one

from raclients.graph.client import GraphQLClient

from sdlon.date_utils import format_date
from sdlon.models import ITUserSystem

QUERY_GET_SD_TO_AD_IT_SYSTEM_UUID = gql(
    """
    query GetItSystems($user_key: [String!]!) {
        itsystems(user_keys: $user_key) {
            objects {
                uuid
            }
        }
    }
"""
)

QUERY_GET_EMPLOYEE_IT_SYSTEMS = gql(
    """
        query GetEmployeeItSystems($uuid: [UUID!]!) {
            employees(uuids: $uuid) {
                objects {
                    current {
                        itusers {
                            itsystem {
                                uuid
                            }
                            user_key
                        }
                    }
                }
            }
        }
    """
)

MUTATION_ADD_IT_SYSTEM_TO_EMPLOYEE = gql(
    """
        mutation AddITSystem($input: ITUserCreateInput!) {
            ituser_create(input: $input) {
                uuid
            }
        }
    """
)


@cache
def get_sd_to_ad_it_system_uuid(
    gql_client: GraphQLClient, it_system_user_key: str
) -> UUID:
    """
    Get the UUID of the "AD-bruger fra SD" IT-system

    Args:
        it_system_user_key: The IT-system user_key
        gql_client: The GraphQL client for calling MO

    Returns:
        UUID of the "AD-bruger fra SD" IT-system
    """
    r = gql_client.execute(
        QUERY_GET_SD_TO_AD_IT_SYSTEM_UUID,
        variable_values={"user_key": it_system_user_key},
    )
    return UUID(one(r["itsystems"]["objects"])["uuid"])


def get_employee_it_systems(
    gql_client: GraphQLClient, employee_uuid: UUID
) -> list[ITUserSystem]:
    """
    Get the IT-systems for an employee
    Args:
        gql_client: The GraphQL client for calling MO
        employee_uuid: The employee UUID

    Returns:
        List of ITUserSystems containing the UUID of the IT-system itself and
        the user key of the IT-user
    """

    r = gql_client.execute(
        QUERY_GET_EMPLOYEE_IT_SYSTEMS, variable_values={"uuid": str(employee_uuid)}
    )

    it_users = one(r["employees"]["objects"])["current"]["itusers"]

    it_user_systems = [
        ITUserSystem(
            uuid=UUID(it_user["itsystem"]["uuid"]), user_key=it_user["user_key"]
        )
        for it_user in it_users
    ]

    return it_user_systems


def add_it_system_to_employee(
    gql_client: GraphQLClient,
    employee_uuid: UUID,
    it_system_uuid: UUID,
    it_system_user_key: str,
) -> None:
    """
    Add the "AD-bruger fra SD" IT-system to a MO employee

    Args:
        gql_client: The GraphQL client for calling MO
        employee_uuid: UUID of the MO employee
        it_system_uuid: UUID of the "AD-bruger fra SD" IT-system
        it_system_user_key: The user_key of the IT-system
    """
    gql_client.execute(
        MUTATION_ADD_IT_SYSTEM_TO_EMPLOYEE,
        variable_values={
            "input": {
                "user_key": it_system_user_key,
                "itsystem": str(it_system_uuid),
                "validity": {"from": format_date(date.today())},
                "person": str(employee_uuid),
            }
        },
    )



import copy
import uuid
from typing import Callable
from typing import Dict
from typing import Optional
from unittest import mock
from unittest import TestCase
from uuid import uuid4

from more_itertools import first_true
from parameterized import parameterized

from .. import ad_life_cycle
from ..ad_exceptions import NoPrimaryEngagementException
from .mocks import MO_CHILD_ORG_UNIT_UUID
from .mocks import MO_ROOT_ORG_UNIT_UUID
from .mocks import MockADParameterReader
from .mocks import MockEmptyADReader
from .mocks import MockLoraCacheDanglingParentUnit
from .mocks import MockLoraCacheEmptyEmployee
from .mocks import MockLoraCacheEmptyUnit
from .mocks import MockLoraCacheExtended
from .mocks import MockLoraCacheParentChildUnit
from .mocks import MockLoraCacheParentUnitUnset
from .test_utils import TestADWriterMixin


BASE_SETTINGS = {"integrations.ad.write.create_user_trees": [MO_ROOT_ORG_UNIT_UUID]}


def mock_find_primary_engagement(eng_org_unit_uuid):
    """Return mock implementation of `MODataSource.find_primary_engagement`"""

    def mock(mo_user_uuid):
        return (
            None,  # = employment_number
            None,  # = title
            eng_org_unit_uuid,
            None,  # = eng_uuid
        )

    return mock


class TestAdLifeCycle(TestCase, TestADWriterMixin):
    """Test `ad_life_cycle`"""

    maxDiff = None

    def setUp(self):
        super().setUp()
        self._setup_adwriter()

    @parameterized.expand(
        [
            # 1. An AD user must be created when the AD is empty.
            # In this test, we mock an empty AD. Therefore we expect exactly
            # one AD write (`New-ADUser`) to be performed.
            (
                # Empty `_get_instance` kwargs
                {},
                # Expected stats
                lambda instance: {
                    "created_users": 1,
                    "users": {instance._prepare_static_person()["uuid"]},
                },
                # Assert a single `New-ADUser` PowerShell command was issued
                {"num_scripts": 1, "expected_script_content": "New-ADUser"},
            ),
            # 2. An AD user must *not* be created when the MO user has no
            # primary engagement.
            # In this test, the AD is empty, but the MO user has no primary
            # engagements. Consequently, no AD users are created.
            (
                # Mock a MO user without a primary engagement
                {
                    "find_primary_engagement": mock.Mock(
                        side_effect=NoPrimaryEngagementException()
                    ),
                },
                # Expected stats
                lambda instance: {
                    "not_in_user_tree": 1,
                    "skipped": instance._get_expected_skipped_entry(
                        "filter_users_outside_unit_tree"
                    ),
                },
                # Assert no AD writes were done
                None,
            ),
            # 3. An AD user must *not* be created when `users_with_engagements`
            # is empty.
            (
                # Mock a MO user without *any* engagements
                {"users_with_engagements": {}},
                # Expected stats
                lambda instance: {
                    "no_active_engagements": 1,
                    "skipped": instance._get_expected_skipped_entry(
                        "filter_user_without_engagements"
                    ),
                },
                # Assert no AD writes were done
                None,
            ),
            # 4. An AD user must *not* be created if the AD user already
            # exists.
            (
                # In this test, the MO user has already been "created" in AD,
                # as the `MockADReader` always returns the same "person"
                # regardless of the CPR number looked up.
                {"reader": MockADParameterReader()},
                # Expected stats
                lambda instance: {
                    "already_in_ad": 1,
                    "skipped": instance._get_expected_skipped_entry(
                        "filter_user_already_in_ad"
                    ),
                },
                # Assert no AD writes were done
                None,
            ),
            # 5. An AD user must *not* be created if at least one "create
            # filter" returns `True` for the corresponding MO user.
            (
                # A "create filter" is a Jinja template evaluating to `True` or
                # `False`. Here is a "create filter" which never returns `True`
                # for our test MO user.
                {"create_filters": ["{{ employee['employment_number'] == '1st' }}"]},
                # Expected stats
                lambda instance: {
                    "create_filtered": 1,
                    "skipped": instance._get_expected_skipped_entry(
                        "create_filters_num_0"
                    ),
                },
                # Assert no AD writes were done
                None,
            ),
            # 6. An AD user must *not* be created if the corresponding MO does not
            # belong to the specified "user trees".
            (
                {
                    # Pretend the AD is empty
                    "reader": {},
                    # Pretend the user tree(s) do not include the MO user,
                    "user_trees": [str(uuid.uuid4())],
                },
                # Expected stats
                lambda instance: {
                    "not_in_user_tree": 1,
                    "skipped": instance._get_expected_skipped_entry(
                        "filter_users_outside_unit_tree"
                    ),
                },
                # Assert no AD writes were done
                None,
            ),
        ]
    )
    def test_create_ad_accounts(
        self,
        instance_kwargs: Dict,
        expected_stats_func: Callable,
        expected_ad_writes: Optional[Dict],
    ):
        instance = self._get_instance(**instance_kwargs)
        stats = instance.create_ad_accounts()
        self._assert_stats_equal(stats, **expected_stats_func(self))
        if expected_ad_writes:
            self._assert_ad_write(**expected_ad_writes)
        else:
            self._assert_no_ad_writes()

    @parameterized.expand(
        [
            # 1. If the AD user exists, and the MO user has *no* engagements,
            # the AD user should be disabled.
            (
                {
                    # Pretend AD user already exists
                    "reader": MockADParameterReader(),
                    # Pretend the MO user has no engagements
                    "users_with_engagements": {},
                },
                # Expected stats
                lambda instance: {
                    "disabled_users": 1,
                    "users": {instance._prepare_static_person()["uuid"]},
                },
                # Assert one AD write is done, which disables the AD user
                {"num_scripts": 1, "expected_script_content": "Disable-ADAccount"},
            ),
            # 2. If the AD user exists, and the MO user has *no* engagements,
            # *but* a "disable filter" filters out the MO user, the AD user
            # should *not* be disabled.
            (
                {
                    # Pretend AD user already exists
                    "reader": MockADParameterReader(),
                    # Pretend the MO user has no engagements
                    "users_with_engagements": {},
                    # Add a "disable filter" which evaluates to `False`, which
                    # means this user should be skipped in
                    # `AdLifeCycle.disable_ad_accounts`.
                    "disable_filters": ["{{ employee['employment_number'] == '1st' }}"],
                },
                # Expected stats
                lambda instance: {
                    "skipped": instance._get_expected_skipped_entry(
                        "disable_filters_num_0"
                    ),
                },
                # Assert no AD writes were done
                None,
            ),
            # 3. If the AD user exists, but the MO user *has* engagements, the
            # AD user should *not* be disabled.
            (
                {
                    # Pretend AD user already exists
                    "reader": MockADParameterReader(),
                    # Pretend the MO user *has* engagements
                    "users_with_engagements": None,
                },
                # Expected stats
                lambda instance: {"disabled_users": 0, "users": set()},
                # Assert no AD writes were done
                None,
            ),
            # 4. If the AD user does not exist, and the MO does not have any
            # engagements, the AD user should *not* be disabled.
            (
                {
                    # Pretend AD does not exist
                    "reader": None,
                    # Pretend the MO user has no engagements
                    "users_with_engagements": {},
                },
                # Expected stats
                lambda instance: {"disabled_users": 0, "users": set()},
                # Assert no AD writes were done
                None,
            ),
            # 5. If the MO user exists and has no engagements, but is not in the
            # "user trees", the corresponding AD user should *not* be disabled.
            (
                {
                    # Pretend AD user already exists
                    "reader": MockADParameterReader(),
                    # Pretend the MO user has no engagements
                    "users_with_engagements": {},
                    # Pretend the user tree(s) do not include the MO user,
                    "user_trees": [str(uuid.uuid4())],
                },
                # Expected stats
                lambda instance: {
                    "not_in_user_tree": 1,
                    "skipped": instance._get_expected_skipped_entry(
                        "filter_users_outside_unit_tree"
                    ),
                },
                # Assert no AD writes were done
                None,
            ),
        ]
    )
    def test_disable_ad_accounts(
        self,
        instance_kwargs: Dict,
        expected_stats_func: Callable,
        expected_ad_writes: Optional[Dict],
    ):
        """An AD user must *not* be disabled if present in
        `users_with_engagements`.
        An AD user can only be disabled if already present in the AD.
        An AD user must *not* be disabled if the MO user exists outside the specified
        'user trees'.
        """
        instance = self._get_instance(**instance_kwargs)
        stats = instance.disable_ad_accounts()
        self._assert_stats_equal(stats, **expected_stats_func(self))
        if expected_ad_writes:
            self._assert_ad_write(**expected_ad_writes)
        else:
            self._assert_no_ad_writes()

    def test_disable_ad_accounts_handles_status(self):
        """If disabling an AD user fails, the status should be recorded."""
        # Get an `AdLifeCycle` instance which is set up to disable one AD user
        instance = self._get_instance(
            reader=MockADParameterReader(), users_with_engagements={}
        )
        # But replace `ADWriter.enable_user` with a function indicating a
        # failed AD update.
        instance.ad_writer.enable_user = lambda **kwargs: (False, "message")
        stats = instance.disable_ad_accounts()
        self._assert_stats_equal(stats, disabled_users=0, critical_errors=1)

    def test_gen_filtered_employees(self):
        """The lazy dicts produced by `_gen_filtered_employees` must be able
        to be evaluated.
        """
        # Consume the lazy dicts produced by `_gen_filtered_employees` to
        # verify their contents.
        instance = self._get_instance()
        consumed = list(instance._gen_filtered_employees())
        self.assertNotEqual(consumed, [])
        for employee, _ad_object in consumed:
            for engagement in employee["engagements"]:
                # Cast `engagement` to dict to force evaluation of its lazy
                # properties.
                self.assertIsInstance(dict(engagement), dict)

    def test_gen_filtered_employees_handles_empty_employees(self):
        instance = self._get_instance(mock_lora_cache_class=MockLoraCacheEmptyEmployee)
        consumed = list(instance._gen_filtered_employees())
        self.assertListEqual(consumed, [])

    @parameterized.expand(
        [
            # 1. Mock employee data from MO API
            (
                # Mock employee data
                dict(uuid=uuid4(), name=("Givenname Middlename", "Surname")),
                # Expected name in log
                "Givenname Middlename Surname",
            ),
            # 2. Mock employee data from LoraCache
            (
                # Mock employee data
                dict(uuid=uuid4(), navn="Givenname Middlename Surname"),
                # Expected name in log
                "Givenname Middlename Surname",
            ),
            # 3. Mock bogus employee data without "name" or "navn" keys
            (
                # Mock employee data
                dict(uuid=uuid4()),
                # Expected name in log
                "unknown",
            ),
        ]
    )
    def test_log_skipped_decorator(self, employee, expected_name):
        instance = self._get_instance()

        @instance.log_skipped("foobar")
        def foobar(tup):
            return False

        # Call decorated function with mock `(employee, ad_user)` tuple
        foobar((employee, None))

        # Expect to find a dictionary mapping employee UUID to employee name
        # under `instance.stats["skipped"]["foobar"]`.
        self.assertDictEqual(
            instance.stats["skipped"]["foobar"],
            {employee["uuid"]: expected_name},
        )

    @parameterized.expand(
        [
            # Case 1: we look for the root unit and find it.
            (
                mock_find_primary_engagement(MO_ROOT_ORG_UNIT_UUID),
                MockLoraCacheExtended,
                True,
            ),
            # Case 2: we look for a child of the root unit, and find it.
            (
                mock_find_primary_engagement(MO_CHILD_ORG_UNIT_UUID),
                MockLoraCacheParentChildUnit,
                True,
            ),
            # Case 3: we look for the root unit, but there are no units at all.
            (
                mock_find_primary_engagement(MO_ROOT_ORG_UNIT_UUID),
                MockLoraCacheEmptyUnit,
                False,
            ),
            # Case 4: we look for a child unit, but the parent-child relation
            # is 'broken', and we can not follow the relationship up to the
            # root node.
            (
                mock_find_primary_engagement(MO_CHILD_ORG_UNIT_UUID),
                MockLoraCacheDanglingParentUnit,
                False,
            ),
            # Case 5: we look for a child unit, but the child node has an unset
            # parent node reference, and we cannot find the root unit.
            (
                mock_find_primary_engagement(MO_CHILD_ORG_UNIT_UUID),
                MockLoraCacheParentUnitUnset,
                False,
            ),
        ]
    )
    def test_find_user_unit_tree(
        self,
        find_primary_engagement: Callable,
        mock_lora_cache_class: Callable,
        expected_result: bool,
    ):
        instance = self._get_instance(
            find_primary_engagement=find_primary_engagement,
            mock_lora_cache_class=mock_lora_cache_class,
        )
        result = instance._find_user_unit_tree(({"uuid": uuid4()}, {}))
        self.assertEqual(result, expected_result)

    def test_preview_command_for_uuid(self):
        with mock.patch("click.echo_via_pager") as mock_echo:
            instance = self._get_instance()
            commands = ad_life_cycle.run_preview_command_for_uuid(instance, "mo_uuid")
            self.assertEqual(mock_echo.call_count, 2)
            self.assertEqual(len(commands), 2)
            self.assertIsInstance(commands[0], str)
            self.assertIsInstance(commands[1], str)

    def test_unhandled_exception(self, *args):
        instance = self._get_instance()
        with mock.patch.object(self.ad_writer, "create_user", side_effect=Exception):
            with self.assertLogs("export") as cm:
                instance.create_ad_accounts()
                self.assertEqual(len(cm.records), 1)
                self.assertRegex(
                    cm.records[0].message,
                    r"Error creating AD user for MO user '.*?': .*",
                )

    def _get_instance(
        self,
        reader=None,
        find_primary_engagement=None,
        users_with_engagements=None,
        create_filters=None,
        disable_filters=None,
        user_trees=None,
        mock_lora_cache_class=MockLoraCacheExtended,
        **kwargs,
    ):
        settings = copy.copy(BASE_SETTINGS)

        if create_filters:
            settings["integrations.ad.lifecycle.create_filters"] = create_filters

        if disable_filters:
            settings["integrations.ad.lifecycle.disable_filters"] = disable_filters

        if user_trees:
            settings["integrations.ad.write.create_user_trees"] = user_trees

        # Replace `find_primary_engagement` with our mocked version
        self.ad_writer.datasource.find_primary_engagement = (
            find_primary_engagement
            or mock_find_primary_engagement(MO_ROOT_ORG_UNIT_UUID)
        )

        load_settings_mock = mock.patch.object(
            ad_life_cycle,
            "load_settings",
            return_value=settings,
        )
        read_settings_mock = mock.patch.object(
            ad_life_cycle,
            "injected_settings",
            return_value=settings,
        )

        lora_cache_mock = mock.patch.object(
            ad_life_cycle,
            "LoraCache",
            new=lambda **kwargs: mock_lora_cache_class(self._prepare_static_person()),
        )

        reader_mock = mock.patch.object(
            ad_life_cycle,
            "ADParameterReader",
            new=lambda: reader or MockEmptyADReader(),
        )

        writer_mock = mock.patch.object(
            ad_life_cycle,
            "ADWriter",
            new=lambda **kwargs: self.ad_writer,
        )

        with load_settings_mock:
            with read_settings_mock:
                with lora_cache_mock:
                    with reader_mock:
                        with writer_mock:
                            instance = ad_life_cycle.AdLifeCycle(**kwargs)
                            # Replace `users_with_engagements` dict attr with our
                            # mocked version.
                            if users_with_engagements is not None:
                                instance.users_with_engagements = {}
                            return instance

    def _assert_stats_equal(self, actual_stats, **expected_stats):
        _keys = {
            "critical_errors",
            "engagement_not_found",
            "created_users",
            "disabled_users",
            "already_in_ad",
            "no_active_engagements",
            "not_in_user_tree",
            "create_filtered",
        }
        _expected_stats = dict.fromkeys(_keys, 0)
        _expected_stats["users"] = set()
        _expected_stats.update(expected_stats)
        self.assertDictEqual(actual_stats, _expected_stats)

    def _assert_no_ad_writes(self):
        self.assertEqual(len(self.ad_writer.scripts), 0)

    def _assert_ad_write(self, num_scripts=None, expected_script_content=None):
        self.assertEqual(len(self.ad_writer.scripts), num_scripts)
        first_true(
            self.ad_writer.scripts,
            pred=lambda script: expected_script_content in script,
        )

    def _get_expected_skipped_entry(self, filtername):
        mo_user = self._prepare_static_person()
        return {filtername: {mo_user["uuid"]: " ".join(mo_user["name"])}}


class _TestableAdLifeCycle(ad_life_cycle.AdLifeCycle):
    def _load_settings(self):
        return BASE_SETTINGS

    def _get_adreader(self):
        return mock.MagicMock()

    def _get_adwriter(self, **kwargs):
        return mock.MagicMock(**kwargs)

    def _update_lora_cache(self, dry_run: bool = True):
        return mock.MagicMock(), mock.MagicMock()


class TestOccupiedNamesCheckFlag(TestCase):
    @parameterized.expand(
        [
            (True,),
            (False,),
        ]
    )
    @mock.patch.object(
        ad_life_cycle,
        "injected_settings",
    )
    def test_skip_occupied_names_check(self, value, injected_settings_mock):
        """If `skip_occupied_names_check` is passed, pass it to `ADWriter`"""
        instance = _TestableAdLifeCycle(skip_occupied_names_check=value)
        self.assertEqual(instance.ad_writer.skip_occupied_names, value)



# SPDX-FileCopyrightText: Magenta ApS
#
# SPDX-License-Identifier: MPL-2.0
"""Integration endpoints."""
import logging
from threading import Lock
from typing import Any

from fastapi import APIRouter
from fastapi import BackgroundTasks
from fastapi import HTTPException
from fastapi import Query
from pydantic import AnyHttpUrl
from pydantic import BaseSettings
from pydantic import Field
from pydantic import SecretStr
from ra_utils.job_settings import JobSettings

from .lora_gql_equivalence_tester import cache_equivalence
from .sql_export import SqlExport

logger = logging.getLogger(__name__)
trigger_router = APIRouter()
# Concurrency lock to ensure that only 1 operation is running at a time
lock_actual = Lock()
lock_historic = Lock()


class DatabaseConfiguration(BaseSettings):  # type: ignore
    class Config:
        frozen = True

    type: str | None
    host: str | None
    db_name: str
    user: str | None
    password: SecretStr | None


class DatabaseSettings(JobSettings):
    class Config:
        frozen = True
        env_nested_delimiter = "__"

    mora_base: AnyHttpUrl = Field("http://mo-service:5000")
    mox_base: AnyHttpUrl = Field("http://mo-service:5000/lora")

    actual_state: DatabaseConfiguration = Field(default_factory=DatabaseConfiguration)
    historic_state: DatabaseConfiguration | None
    log_overlapping_aak: bool = False
    use_new_cache: bool = False
    primary_manager_responsibility: str | None = None

    def to_old_settings(self) -> dict[str, Any]:
        """Convert our DatabaseSettings to a settings.json format.

        This serves to implement the adapter pattern, adapting from pydantic and its
        corresponding 12-factor configuration paradigm with environmental variables, to
        the current functionality of the program, based on the settings format from
        settings.json.

        Eventually the entire settings-processing within the program should be
        rewritten with a process similar to what has been done for the SD integration,
        but it was out of scope for the change when this code was introduces.
        """

        def secret_val_or_none(value: SecretStr | None) -> str | None:
            if value is None:
                return None
            return value.get_secret_value()

        settings = {
            "mora.base": self.mora_base,
            "mox.base": self.mox_base,
            "exporters.actual_state.type": self.actual_state.type,
            "exporters.actual_state.host": self.actual_state.host,
            "exporters.actual_state.db_name": self.actual_state.db_name,
            "exporters.actual_state.user": self.actual_state.user,
            "exporters.actual_state.password": secret_val_or_none(
                self.actual_state.password
            ),
            "primary_manager_responsibility": self.primary_manager_responsibility,
            "use_new_cache": self.use_new_cache,
        }
        if self.historic_state is not None:
            settings.update(
                {
                    "exporters.actual_state_historic.type": self.historic_state.type
                    or self.actual_state.type,
                    "exporters.actual_state_historic.host": self.historic_state.host
                    or self.actual_state.host,
                    "exporters.actual_state_historic.db_name": self.historic_state.db_name
                    or self.actual_state.db_name,
                    "exporters.actual_state_historic.user": self.historic_state.user
                    or self.actual_state.user,
                    "exporters.actual_state_historic.password": secret_val_or_none(
                        self.historic_state.password
                    )
                    or secret_val_or_none(self.actual_state.password),
                    "primary_manager_responsibility": self.primary_manager_responsibility,
                    "use_new_cache": self.use_new_cache,
                }
            )
        return settings


def refresh_db(
    resolve_dar: bool, historic: bool, read_from_cache: bool, lock: Lock
) -> None:
    try:
        logger.info("*SQL export started*")
        database_settings = DatabaseSettings()

        sql_export = SqlExport(
            force_sqlite=False,
            historic=historic,
            settings=database_settings.to_old_settings(),
        )
        sql_export.perform_export(
            resolve_dar=resolve_dar,
            use_pickle=read_from_cache,
        )

        sql_export.swap_tables()
        logger.info("*SQL export ended*")
    finally:
        # Another operation can start now
        lock.release()


@trigger_router.post("/trigger_cache_equivalence")
async def trigger_cache_equivalence(
    background_tasks: BackgroundTasks,
) -> dict[str, str]:
    background_tasks.add_task(cache_equivalence)
    return {"triggered": "OK"}


@trigger_router.post("/trigger")
def trigger(
    background_tasks: BackgroundTasks,
    resolve_dar: bool = Query(False),
    historic: bool = Query(False),
    read_from_cache: bool = Query(False),
) -> dict[str, str]:
    if historic:
        lock = lock_historic
    else:
        lock = lock_actual
    acquired = lock.acquire(blocking=False)
    if not acquired:
        settings = DatabaseSettings()
        if settings.log_overlapping_aak:
            sql_export = SqlExport(
                force_sqlite=False,
                historic=historic,
                settings=settings.to_old_settings(),
            )
            sql_export.log_overlapping_runs_aak()
        raise HTTPException(409, "Already running")

    background_tasks.add_task(refresh_db, resolve_dar, historic, read_from_cache, lock)
    return {"detail": "Triggered"}



from typing import Dict
from typing import Optional

import click
import requests
from click_option_group import MutuallyExclusiveOptionGroup
from click_option_group import optgroup
from more_itertools import pairwise
from more_itertools import prepend
from ra_utils.load_settings import load_settings

from integrations.ad_integration import ad_reader
from integrations.opus import opus_helpers
from integrations.opus.opus_diff_import import import_one
from tools.data_fixers.class_tools import find_duplicates_classes
from tools.subtreedeleter import subtreedeleter_helper


def truncate_db(MOX_BASE: str = "http://localhost:5000/lora") -> None:
    r = requests.get(MOX_BASE + "/db/truncate")
    r.raise_for_status()


def prepare_re_import(
    settings: Optional[Dict] = None,
    opus_uuid: Optional[str] = None,
    connections: int = 4,
) -> None:
    """Remove all opus-units from MO"""
    settings = settings or load_settings()
    mox_base = settings.get("mox.base")
    if opus_uuid:
        session = requests.session()
        dub = find_duplicates_classes(session=session, mox_base=mox_base)
        if dub:
            raise Exception(
                "There are duplicate classes, remove them with tools/data_fixers/remove_duplicate_classes.py --delete"
            )
        subtreedeleter_helper(
            opus_uuid,
            delete_functions=True,
            keep_functions=["KLE", "Relateret Enhed"],
            connections=connections,
        )


def import_opus(
    ad_reader=None,
    import_all: bool = False,
    import_last=False,
    opus_id=None,
    rundb_write: bool = True,
) -> None:
    """Import one or all files from opus even if no previous files have been imported"""
    settings = load_settings()
    filter_ids = settings.get("integrations.opus.units.filter_ids", [])
    dumps = opus_helpers.read_available_dumps()

    all_dates = dumps.keys()
    # Default is read first file only
    export_dates = [min(all_dates)]
    if import_last:
        export_dates = [max(all_dates)]
    elif import_all:
        export_dates = sorted(all_dates)

    all_export_dates = prepend(None, export_dates)
    date_pairs = pairwise(all_export_dates)
    for date1, date2 in date_pairs:
        import_one(
            ad_reader,
            date2,
            date1,
            dumps,
            filter_ids,
            opus_id=opus_id,
            rundb_write=rundb_write,
        )


@click.command()
@optgroup.group(
    "Import Amount",
    cls=MutuallyExclusiveOptionGroup,
    help="Default: Imports first file",
)
@optgroup.option(
    "--import-all",
    is_flag=True,
    help="Import all opus files.",
)
@optgroup.option(
    "--import-last",
    is_flag=True,
    help="Import only latest file.",
)
@optgroup.group(
    "Clear data", cls=MutuallyExclusiveOptionGroup, help="Default: No data deleted"
)
@optgroup.option(
    "--delete-opus",
    is_flag=True,
    help="Delete Opus subtree. Deletes all units, engagements etc, but not KLE and related units",
)
@click.option(
    "--connections",
    default=4,
    help="The amount of concurrent requests made to OS2mo. Only applies to '--delete-opus'",
)
@click.option("--new-rundb", is_flag=True, help="Initialize new run-db")
@click.option("--use-ad", is_flag=True, help="Read from AD")
def clear_and_reload(
    import_all: bool,
    import_last: bool,
    delete_opus: bool,
    new_rundb: bool,
    use_ad: bool,
    connections: int,
) -> None:
    """Tool for reimporting opus files.

    This tool will load the first opus-file into MO if no inputs are given.
    If the opus organisation allready exists, use either --delete-opus or --truncate.
    --delete-opus will use an uuid genereted from the first unit of the first opus-file and delete all units under this unit, and all relations to thoose units except KLE and related units.
    --truncate will truncate the database entirely.
    Add the --use-ad flag to connect to AD when reading users.
    """
    settings = load_settings()
    if new_rundb:
        opus_helpers.initialize_db(settings["integrations.opus.import.run_db"])

    opus_uuid = str(opus_helpers.find_opus_root_unit_uuid()) if delete_opus else None
    prepare_re_import(
        settings=settings,
        opus_uuid=opus_uuid,
        connections=connections,
    )
    AD = None
    if use_ad:
        AD = ad_reader.ADParameterReader()
        AD.cache_all(print_progress=True)
    import_opus(ad_reader=AD, import_all=import_all, import_last=import_last)


if __name__ == "__main__":
    clear_and_reload()



import click
from more_itertools import first
from more_itertools import only
from os2mo_helpers.mora_helpers import MoraHelper
from ra_utils.load_settings import load_setting

from integrations.ad_integration import ad_reader
from integrations.opus import opus_helpers
from integrations.opus.clear_and_import_opus import import_opus
from integrations.opus.opus_file_reader import get_opus_filereader
from tools.data_fixers.remove_from_lora import delete_object_and_orgfuncs


def find_type(opus_id, full_history):
    """Check if the object with the given id is a unit or an employee."""
    dumps = get_opus_filereader().list_opus_files()
    # Search in newest file first
    opus_files = sorted(dumps, reverse=True)
    if not full_history:
        opus_files = [first(opus_files)]
    for f in opus_files:
        units, employees = opus_helpers.parser(dumps[f], opus_id=opus_id)
        employees, terminated_employees = opus_helpers.split_employees_leaves(employees)
        employees = list(employees)
        if units:
            return "organisationenhed", only(units)
        elif employees:
            return "bruger", only(employees)

    terminated_employees = list(terminated_employees)
    if terminated_employees:
        msg = "Employee was terminated, try --full-history"
    else:
        msg = f"No object with {opus_id=} was found."
    raise ValueError(msg)


@click.command()
@click.option("--mox-base", default=load_setting("mox.base"))
@click.option("--mora-base", default=load_setting("mora.base"))
@click.option("--delete", is_flag=True)
@click.option("--full-history", is_flag=True)
@click.option("--opus-id", type=int, required=True)
@click.option("--use-ad", is_flag=True, help="Read from AD")
@click.option("--dry-run", is_flag=True)
def cli(mox_base, mora_base, delete, full_history, opus_id, use_ad, dry_run):
    """Reimport object from opus with given opus-ID to MO
    Optionally deletes the object and all related orgfuncs directly from Lora.
    Defaults to reading latest file only, but supports reading full history
    """
    helper = MoraHelper(hostname=mora_base)
    object_type, obj = find_type(opus_id, full_history)
    if object_type == "bruger":
        cpr = opus_helpers.read_cpr(obj)
        user = helper.read_user(user_cpr=cpr)
        uuid = user["uuid"] if user else None
    else:
        uuid = opus_helpers.generate_uuid(obj["@id"])

    if delete and uuid and not dry_run:
        delete_object_and_orgfuncs(uuid, mox_base, object_type)
    if dry_run:
        click.echo(
            f"Dry-run: {'Delete and reimport' if delete else 'Reimport'} '{object_type}' with {uuid=}"
        )
    else:
        AD = ad_reader.ADParameterReader() if use_ad else None
        import_opus(
            ad_reader=AD,
            import_all=full_history,
            import_last=not full_history,
            opus_id=opus_id,
            rundb_write=False,
        )


if __name__ == "__main__":
    cli()



from collections import OrderedDict
from copy import deepcopy
from unittest.mock import patch, MagicMock

import pytest
from parameterized import parameterized

from sdlon.config import ChangedAtSettings
from sdlon.skip import cpr_env_filter, skip_job_position_id
from .test_config import DEFAULT_CHANGED_AT_SETTINGS


class TestCprEnvFilter:
    @parameterized.expand(
        [
            (True, [], True),
            (True, ["5555555555"], False),
            (False, [], False),
            (False, ["5555555555"], True),
        ]
    )
    @patch("sdlon.skip.get_changed_at_settings")
    def test_return_true_for_exclude_mode_when_cpr_not_in_list(
        self, exclude_mode, cprs, expected, mock
    ):
        settings_dict = deepcopy(DEFAULT_CHANGED_AT_SETTINGS)
        settings_dict.update({"sd_exclude_cprs_mode": exclude_mode, "sd_cprs": cprs})
        mock.return_value = ChangedAtSettings.parse_obj(settings_dict)

        entity = {"PersonCivilRegistrationIdentifier": "5555555555"}

        assert cpr_env_filter(entity) == expected


@pytest.mark.parametrize(
    "profession,job_pos_ids_to_skip,expected",
    [
        ({}, ["1", "2", "3"], False),
        ({"JobPositionIdentifier": "4"}, ["1", "2", "3"], False),
        ({"JobPositionIdentifier": "1"}, ["1", "2", "3"], True),
        ({"JobPositionIdentifier": "1"}, [], False),
    ],
)
def test_profession_job_position_id_filter(
    profession: dict,
    job_pos_ids_to_skip: list[str],
    expected: bool,
):
    """
    Test sdlon.skip.skip_job_position_id.

    Args:
        profession: the <Profession> part of the <Employment> of the SD payload
        expected: the expected return value of skip_job_position_id
        job_pos_ids_to_skip: the JobPositionIdentifiers to skip
    """

    # Act
    result = skip_job_position_id(OrderedDict(profession), job_pos_ids_to_skip)

    # Assert
    assert result is expected



import asyncio
from collections import ChainMap
from functools import partial
from operator import itemgetter
from typing import List
from typing import Optional
from typing import Tuple

from aiohttp import ClientSession
from aiohttp import TCPConnector
from more_itertools import chunked
from more_itertools import unzip
from ra_utils.async_to_sync import async_to_sync


def ensure_session(func):
    async def _decorator(*args, **kwargs):
        if "client" in kwargs:
            return await func(*args, **kwargs)
        else:
            # DAWA only accepts:
            # * 30 requests per second per IP
            # * 10 concurrent connections per IP
            # Thus we limit our connections to 10 here.
            connector = TCPConnector(limit=10)
            async with ClientSession(connector=connector) as client:
                return await func(*args, **kwargs, client=client)

    return _decorator


@ensure_session
async def dar_fetch_non_chunked(uuids, addrtype, client=None):
    """Lookup uuids in DAR (without chunking).

    Args:
        uuids: List of DAR UUIDs to lookup.
        addr_type: The address type to lookup.
        client (optional): aiohttp.ClientSession to use for connecting.

    Returns:
        (dict, set):
            dict: Map from UUID to DAR reply.
            set: Set of UUIDs of entries which were not found.
    """
    url = "https://dawa.aws.dk/" + addrtype
    params = {"id": "|".join(uuids), "struktur": "mini"}

    async with client.get(url, params=params) as response:
        response.raise_for_status()
        body = await response.json()

        result = {addr["id"]: addr for addr in body}

        found_uuids = set(map(itemgetter("id"), body))
        missing = set(uuids) - found_uuids

        return result, missing


@ensure_session
async def dar_fetch_chunked(uuids, addrtype, chunk_size, client=None):
    """Lookup uuids in DAR (chunked).

    Args:
        uuids: List of DAR UUIDs.
        addr_type: The address type to lookup.
        chunk_size: Number of UUIDs per block, sent to DAR.
        client (optional): aiohttp.ClientSession to use for connecting.

    Returns:
        (dict, set):
            dict: Map from UUID to DAR reply.
            set: Set of UUIDs of entries which were not found.
    """

    def create_task(uuid_chunk):
        return asyncio.ensure_future(
            dar_fetch_non_chunked(uuid_chunk, addrtype=addrtype, client=client)
        )

    # Chunk our UUIDs into blocks of chunk_size
    uuid_chunks = chunked(uuids, chunk_size)
    # Convert chunks into a list of asyncio.tasks
    tasks = list(map(create_task, uuid_chunks))
    # Here 'result' is a list of tuples (dict, set) => (result, missing)
    result = await asyncio.gather(*tasks)
    # First we unzip 'result' to get a list of results and a list of missing
    result_dicts, missing_sets = unzip(result)
    # Then we union the dicts and sets before returning
    combined_result = dict(ChainMap(*result_dicts))
    combined_missing = set.union(*missing_sets)
    return combined_result, combined_missing


@ensure_session
async def dar_fetch(uuids, addrtype="adresser", chunk_size=150, client=None):
    """Lookup uuids in DAR (chunked if required).

    Args:
        uuids: List of DAR UUIDs.
        addr_type: The address type to lookup.
        chunk_size: Number of UUIDs per block, sent to DAR.
        client (optional): aiohttp.ClientSession to use for connecting.

    Returns:
        (dict, set):
            dict: Map from UUID to DAR reply.
            set: Set of UUIDs of entries which were not found.
    """
    num_uuids = len(uuids)
    if num_uuids == 0:
        return dict(), set()
    elif num_uuids <= 150:
        return await dar_fetch_non_chunked(uuids, addrtype, client=client)
    else:
        return await dar_fetch_chunked(uuids, addrtype, chunk_size, client=client)


@async_to_sync
async def sync_dar_fetch(uuids, addrtype="adresser", chunk_size=150):
    """Syncronized version of dar_fetch."""
    return await dar_fetch(uuids, addrtype, chunk_size)


@ensure_session
async def dar_datavask(
    address: str, client: ClientSession
) -> Tuple[str, Optional[dict]]:
    """
    Perform search in DAR using the 'datavask' API
    """
    url = "https://api.dataforsyningen.dk/datavask/adresser"
    params = {"struktur": "mini", "betegnelse": address}
    async with client.get(url, params=params) as response:
        response.raise_for_status()
        result = await response.json()
        # A and B are safe matches, and will only have one result in the array
        # https://dawadocs.dataforsyningen.dk/dok/api/adresse#datavask
        if result["kategori"] in ["A", "B"]:
            return address, result["resultater"][0]["adresse"]["id"]

    return address, None


@ensure_session
async def dar_datavask_multiple(
    addresses: List[str], client: ClientSession
) -> List[Tuple[str, Optional[dict]]]:
    """Perform search in DAR on multiple address strings"""
    tasks = map(partial(dar_datavask, client=client), addresses)
    results = await asyncio.gather(*tasks)
    return results



# Update all engagements with the correct value of "Timelønnet" or
# "Månedslønnet" according to the value of the environment variable
# SD_MONTHLY_HOURLY_DIVIDE. If the SD EmploymentIdentifier (user_key of the
# employment) is above the value set in this ENV, they should be "Timelønnet"
# or else they should be "Månedslønnet"
from datetime import date
from enum import Enum
from uuid import UUID

import click
from gql import gql
from more_itertools import one
from pydantic import BaseModel
from raclients.graph.client import GraphQLClient

from sdlon.date_utils import format_date, parse_datetime
from sdlon.graphql import get_mo_client


class SalaryType(Enum):
    HOURLY = "timeløn"
    MONTHLY = "månedsløn"


class Engagement(BaseModel):
    eng_uuid: UUID
    user_key: str
    eng_type_uuid: UUID
    from_date: date


def get_eng_type_uuid(
        gql_client: GraphQLClient, salary_type: SalaryType
) -> UUID:
    """
    Get the UUID of the "Timelønnet" engagement type
    """
    query = gql(
        """
            query GetEngagementTypeUUID($user_key: [String!]!) {
              classes(user_keys: $user_key) {
                objects {
                  user_key
                  type
                  name
                  uuid
                }
              }
            }
        """
    )

    r = gql_client.execute(query, variable_values={"user_key": salary_type.value})

    return UUID(one(r["classes"]["objects"])["uuid"])


def get_engagement_user_key_and_type(
        gql_client: GraphQLClient
) -> list[Engagement]:
    """
    Get all engagements (user_keys and engagement_types)

    Args:
        gql_client: The GraphQL client

    Returns:
        List of Engagements
    """

    query = gql(
        """
            query GetEngagementAndType {
              engagements {
                objects {
                  objects {
                    uuid
                    user_key
                    engagement_type {
                      uuid
                      user_key
                      name
                    }
                    validity {
                      from
                    }
                  }
                }
              }
            }
        """
    )

    r = gql_client.execute(query)
    objects = r["engagements"]["objects"]

    return [
        Engagement(
            eng_uuid=UUID(one(obj["objects"])["uuid"]),
            user_key=one(obj["objects"])["user_key"],
            eng_type_uuid=UUID(one(obj["objects"])["engagement_type"]["uuid"]),
            from_date=parse_datetime(one(obj["objects"])["validity"]["from"][:10]).date()
        )
        for obj in objects
    ]


def update_engagement_type(
        gql_client: GraphQLClient, eng_uuid: UUID, eng_type_uuid: UUID, from_date: date
) -> None:
    """
    Update the engagement type of an engagement from a given date.

    Args:
        eng_uuid: The engagement UUID
        eng_type_uuid: The UUID of the engagement type
        from_date: The date from which the change should take effect
    """

    mutation = gql(
        """
            mutation UpdateEngagement($input: EngagementUpdateInput!) {
                engagement_update(input: $input) {
                    uuid
                }
            }
        """
    )

    gql_client.execute(mutation, variable_values={
        "input": {
            "uuid": str(eng_uuid),
            "validity": {"from": format_date(from_date)},
            "engagement_type": str(eng_type_uuid)
        }
    })


@click.command()
@click.option(
    "--auth-server",
    "auth_server",
    type=click.STRING,
    default="http://localhost:8090/auth",
    help="Keycloak auth server URL"
)
@click.option(
    "--client-id",
    "client_id",
    type=click.STRING,
    default="dipex",
    help="Keycloak client id"
)
@click.option(
    "--client-secret",
    "client_secret",
    type=click.STRING,
    required=True,
    help="Keycloak client secret for the DIPEX client"
)
@click.option(
    "--mo-base-url",
    "mo_base_url",
    type=click.STRING,
    default="http://localhost:5000",
    help="Base URL for calling MO"
)
@click.option(
    "--monthly-hourly-divide",
    type=click.INT,
    required=True,
    help="The new SD_MONTHLY_HOURLY_DIVIDE"
)
@click.option(
    "--dry-run",
    "dry_run",
    is_flag=True,
    help="Do not perform any changes in MO"
)
def main(
        auth_server: str,
        client_id: str,
        client_secret: str,
        mo_base_url: str,
        monthly_hourly_divide: int,
        dry_run: bool
):
    gql_client = get_mo_client(
        auth_server, client_id, client_secret, mo_base_url, 5
    )

    hourly_eng_type_uuid = get_eng_type_uuid(gql_client, SalaryType.HOURLY)
    monthly_eng_type_uuid = get_eng_type_uuid(gql_client, SalaryType.MONTHLY)

    engagements = get_engagement_user_key_and_type(gql_client)

    hourly_engagements_to_fix = [
        eng for eng in engagements
        if eng.eng_type_uuid == hourly_eng_type_uuid and
           int(eng.user_key) < monthly_hourly_divide
    ]
    print("Number of hourly engagements to fix:", len(hourly_engagements_to_fix))
    for eng in hourly_engagements_to_fix:
        if not dry_run:
            update_engagement_type(
                gql_client, eng.eng_uuid, monthly_eng_type_uuid, eng.from_date)
        print(f"Updated engagement {eng.eng_uuid} from hourly to monthly from {format_date(eng.from_date)}")

    monthly_engagements_to_fix = [
        eng for eng in engagements
        if eng.eng_type_uuid == monthly_eng_type_uuid and
           int(eng.user_key) >= monthly_hourly_divide
    ]
    print("Number of monthly engagements to fix:", len(monthly_engagements_to_fix))
    for eng in monthly_engagements_to_fix:
        if not dry_run:
            update_engagement_type(
                gql_client, eng.eng_uuid, hourly_eng_type_uuid, eng.from_date)
        print(f"Updated engagement {eng.eng_uuid} from monthly to hourly from {format_date(eng.from_date)}")


if __name__ == "__main__":
    main()



from unittest.mock import MagicMock
from uuid import UUID

from sdlon.employees import get_employee, QUERY_GET_EMPLOYEE
from sdlon.models import MOBasePerson
from tests.fixtures import mock_graphql_client


def test_get_employee(mock_graphql_client: MagicMock):
    # Arrange
    mock_execute = MagicMock(
        return_value={
            "employees": {
                "objects": [
                    {
                        "current": {
                            "name": "Solveig Kuhlenhenke",
                            "givenname": "Solveig",
                            "surname": "Kuhlenhenke",
                            "uuid": "23d2dfc7-6ceb-47cf-97ed-db6beadcb09b",
                        }
                    }
                ]
            }
        }
    )
    mock_graphql_client.execute = mock_execute

    # Act
    employee = get_employee(mock_graphql_client, "1111111112")

    # Assert
    mock_execute.assert_called_once_with(
        QUERY_GET_EMPLOYEE, variable_values={"cpr": "1111111112"}
    )
    assert employee == MOBasePerson(
        cpr="1111111112",
        givenname="Solveig",
        surname="Kuhlenhenke",
        name="Solveig Kuhlenhenke",
        uuid=UUID("23d2dfc7-6ceb-47cf-97ed-db6beadcb09b"),
    )


def test_get_employee_none(mock_graphql_client: MagicMock):
    # Arrange
    mock_execute = MagicMock(return_value={"employees": {"objects": []}})
    mock_graphql_client.execute = mock_execute

    # Act
    employee = get_employee(mock_graphql_client, "1111111112")

    # Assert
    mock_execute.assert_called_once_with(
        QUERY_GET_EMPLOYEE, variable_values={"cpr": "1111111112"}
    )
    assert employee is None



#!/usr/bin/env python3
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
import datetime
import logging
import sys
import uuid
from operator import itemgetter
from typing import Any, OrderedDict
from typing import Dict
from typing import Optional

import click
from anytree import Node
from integrations import dawa_helper
from integrations.ad_integration import ad_reader
from os2mo_data_import import ImportHelper
from os2mo_helpers.mora_helpers import MoraHelper

from .config import get_importer_settings
from .config import ImporterSettings
from .date_utils import format_date, date_to_datetime
from .date_utils import get_employment_datetimes
from .date_utils import parse_datetime
from .models import JobFunction
from .sd_common import calc_employment_id
from .sd_common import EmploymentStatus
from .sd_common import ensure_list
from .sd_common import generate_uuid
from .sd_common import read_employment_at
from .sd_common import sd_lookup
from .skip import skip_fictional_users


LOG_LEVEL = logging.DEBUG

HISTORIC = "historic"

logger = logging.getLogger("sdImport")


class SdImport:
    # XXX: This does really expensive calls against SD in __init__, caution!

    def __init__(
        self,
        importer,
        settings: ImporterSettings,
        ad_info=None,
        org_only=False,
        org_id_prefix=None,
        employee_mapping=None,
    ):
        employee_mapping = employee_mapping or {}

        self.settings = settings

        self.base_url = "https://service.sd.dk/sdws/"
        self.address_errors: Dict[str, Dict] = {}

        self.importer = importer

        self.org_name = self.settings.municipality_name
        self.sd_institution_uuid = self._get_institution()

        self.importer.add_organisation(
            identifier=self.org_name,
            user_key=self.org_name,
            municipality_code=self.settings.municipality_code,
            uuid=str(self.sd_institution_uuid),
        )

        self.org_id_prefix = org_id_prefix

        self.import_date = self.settings.sd_global_from_date.strftime("%d.%m.%Y")
        # List of job_functions that should be ignored.
        self.skip_job_functions = self.settings.sd_skip_employment_types

        # Whether to create SD-medarbejder associations
        self.create_associations = self.settings.sd_importer_create_associations

        # Whether to import email addresses for organisations
        self.create_email_addresses = self.settings.sd_importer_create_email_addresses

        # Whether to use <Employment><EmploymentDate> as engagement start date instead
        # of <Employment><EmploymentStatus><ActivationDate>.
        self.employment_date_as_engagement_start_date = (
            self.settings.sd_importer_employment_date_as_engagement_start_date
        )

        self.historic_org_unit_uuid = str(uuid.uuid4())

        # CPR indexed dictionary of AD users
        self.ad_people: Dict[str, Dict] = {}
        self.employee_forced_uuids = employee_mapping
        self.ad_reader = None
        if isinstance(ad_info, ad_reader.ADParameterReader):
            self.ad_reader = ad_info
            self.importer.new_itsystem(identifier="AD", system_name="Active Directory")
            self.ad_reader.cache_all()

        if self.settings.sd_phone_number_id_for_ad_creation:
            self.importer.new_itsystem(
                identifier="AD-bruger fra SD", system_name="AD-bruger fra SD"
            )

        self.nodes: Dict[str, Dict] = {}  # Will be populated when org-tree is created

        self.org_only = org_only
        if not org_only:
            self.add_people()

        self.info = self._read_department_info()

        self._add_classes()

    def _add_classes(self):
        # Format is facet -> list of entries
        # Entries are tuples on the format '(klasse_id, klasse, scope)'
        classes = {
            "manager_type": [
                ("leder_type", "Leder"),
            ],
            "org_unit_type": [
                ("Enhed", "Enhed"),
            ],
            "org_unit_address_type": [
                ("Pnummer", "Pnummer", "PNUMBER"),
                ("AddressMailUnit", "Postadresse", "DAR"),
                ("AdresseReturUnit", "Returadresse", "DAR"),
                ("AdresseHenvendelseUnit", "Henvendelsessted", "DAR"),
                ("PhoneUnit", "Telefon", "PHONE"),
                ("EmailUnit", "Email", "EMAIL"),
            ],
            "employee_address_type": [
                ("AdressePostEmployee", "Postadresse", "DAR"),
                ("PhoneEmployee", "Telefon", "PHONE"),
                ("MobilePhoneEmployee", "Mobiltelefon", "PHONE"),
                ("LocationEmployee", "Lokation", "TEXT"),
                ("EmailEmployee", "Email", "EMAIL"),
            ],
            "leave_type": [
                ("Orlov", "Orlov"),
            ],
            "engagement_job_function": [(HISTORIC, "Ukendt")],
            "engagement_type": [
                ("månedsløn", "Medarbejder (månedsløn)"),
                ("timeløn", "Medarbejder (timeløn)"),
                ("historisk", "Historisk ansættelse"),
            ],
            "primary_type": [
                ("Ansat", "Ansat", "3000"),
                ("status0", "Ansat - Ikke i løn", "1000"),
                ("non-primary", "Ikke-primær ansættelse", "0"),
                ("explicitly-primary", "Manuelt primær ansættelse", "5000"),
            ],
            "association_type": [
                ("SD-medarbejder", "SD-medarbejder"),
            ],
            "visibility": [
                ("Ekstern", "Må vises eksternt", "PUBLIC"),
                ("Intern", "Må vises internt", "INTERNAL"),
                ("Hemmelig", "Hemmelig", "SECRET"),
            ],
            "manager_level": [
                # XXX: Why 1040, 1035 and 1030?
                ("manager_1040", "Leder"),
                ("manager_1035", "Chef"),
                ("manager_1030", "Direktør"),
            ],
            "responsibility": [
                ("Lederansvar", "Lederansvar"),
            ],
        }

        for facet, klasses in classes.items():
            for klass in klasses:
                if len(klass) == 3:
                    klasse_id, klasse, scope = klass
                else:
                    klasse_id, klasse = klass
                    scope = "TEXT"
                self._add_klasse(klasse_id, klasse, facet, scope)

    def _update_ad_map(self, cpr):
        logger.debug("_update_ad_map called")
        self.ad_people[cpr] = {}
        if self.ad_reader:
            response = self.ad_reader.read_user(cpr=cpr, cache_only=True)
            if response:
                self.ad_people[cpr] = response

    def _read_department_info(self):
        """Load all department details and store for later user."""
        department_info = {}

        params = {
            "ActivationDate": self.import_date,
            "DeactivationDate": self.import_date,
            "ContactInformationIndicator": "true",
            "DepartmentNameIndicator": "true",
            "PostalAddressIndicator": "true",
            "ProductionUnitIndicator": "true",
            "UUIDIndicator": "true",
            "EmploymentDepartmentIndicator": "false",
        }
        departments = sd_lookup(
            "GetDepartment20111201", settings=self.settings, params=params
        )

        for department in departments["Department"]:
            uuid = department["DepartmentUUIDIdentifier"]
            if self.org_id_prefix:
                uuid = generate_uuid(uuid, self.org_id_prefix, self.org_name)

            department_info[uuid] = department
            unit_level = department["DepartmentLevelIdentifier"]
            if not self.importer.check_if_exists("klasse", unit_level):
                self._add_klasse(unit_level, unit_level, "org_unit_level", scope="TEXT")
        return department_info

    def _add_klasse(self, klasse_id, klasse, facet, scope="TEXT"):
        if isinstance(klasse_id, str):
            klasse_id = klasse_id.replace("&", "_")
        if not self.importer.check_if_exists("klasse", klasse_id):
            klasse_uuid = generate_uuid(klasse_id, self.org_id_prefix, self.org_name)
            self.importer.add_klasse(
                identifier=klasse_id,
                uuid=klasse_uuid,
                facet_type_ref=facet,
                user_key=str(klasse_id),
                scope=scope,
                title=klasse,
            )
        return klasse_id

    def _check_subtree(self, department, sub_tree):
        """Check if a department is member of given sub-tree"""
        while "DepartmentReference" in department:
            department = department["DepartmentReference"]
            dep_uuid = department["DepartmentUUIDIdentifier"]
            if self.org_id_prefix:
                dep_uuid = generate_uuid(dep_uuid, self.org_id_prefix)
            if dep_uuid == sub_tree:
                return True
        return False

    def _add_sd_department(
        self, department, contains_subunits=False, sub_tree=None, super_unit=None
    ):
        """
        Add add a deparment to MO. If the unit has parents, these will
        also be added
        :param department: The SD-department, including parent units.
        :param contains_subunits: True if the unit has sub-units.
        """
        ou_level = department["DepartmentLevelIdentifier"]
        if not self.org_id_prefix:
            unit_id = department["DepartmentUUIDIdentifier"]
            user_key = department["DepartmentIdentifier"]
        else:
            unit_id = generate_uuid(
                department["DepartmentUUIDIdentifier"], self.org_id_prefix
            )
            user_key = self.org_id_prefix + "_" + department["DepartmentIdentifier"]

        # parent_uuid = None
        parent_uuid = super_unit

        # If contain_subunits is true, this sub tree is a valid member
        import_unit = contains_subunits
        if "DepartmentReference" in department:
            if self._check_subtree(department, sub_tree):
                import_unit = True

            parent_uuid = department["DepartmentReference"]["DepartmentUUIDIdentifier"]
            if self.org_id_prefix:
                parent_uuid = self._generate_uuid(parent_uuid, self.org_id_prefix)
        else:
            import_unit = unit_id == sub_tree

        if not import_unit and sub_tree is not None:
            return

        info = self.info[unit_id]
        assert info["DepartmentLevelIdentifier"] == ou_level
        logger.debug("Add unit: {}".format(unit_id))
        if (
            (not contains_subunits)
            and (parent_uuid is super_unit)
            and self.importer.check_if_exists("organisation_unit", "OrphanUnits")
        ):
            parent_uuid = "OrphanUnits"

        date_from = info["ActivationDate"]
        # No units have termination dates: date_to is None
        if self.importer.check_if_exists("organisation_unit", unit_id):
            return
        else:
            self.importer.add_organisation_unit(
                identifier=unit_id,
                name=info["DepartmentName"],
                user_key=user_key,
                org_unit_level_ref=ou_level,
                type_ref="Enhed",
                date_from=date_from,
                uuid=unit_id,
                date_to=None,
                parent_ref=parent_uuid,
            )

        def import_email_addresses(info: Dict[str, Any]) -> None:
            if "ContactInformation" not in info:
                return
            if "EmailAddressIdentifier" not in info["ContactInformation"]:
                return

            emails = info["ContactInformation"]["EmailAddressIdentifier"]
            # filter empty entities
            emails = filter(lambda email: email.find("Empty") == -1, emails)
            for email in emails:
                self.importer.add_address_type(
                    organisation_unit=unit_id,
                    type_ref="EmailUnit",
                    value=email,
                    date_from=date_from,
                )

        def import_pnumber(info: Dict[str, Any]) -> None:
            if "ProductionUnitIdentifier" not in info:
                return
            self.importer.add_address_type(
                organisation_unit=unit_id,
                type_ref="Pnummer",
                value=info["ProductionUnitIdentifier"],
                date_from=date_from,
            )

        def import_postal_address(info: Dict[str, Any]) -> None:
            if "PostalAddress" not in info:
                return
            postal_address = info["PostalAddress"]

            if "StandardAddressIdentifier" not in postal_address:
                return
            address_string = postal_address["StandardAddressIdentifier"]

            if "PostalCode" not in postal_address:
                return
            zip_code = postal_address["PostalCode"]

            logger.debug("Look in Dawa: {}".format(address_string))
            dar_uuid = dawa_helper.dawa_lookup(address_string, zip_code)
            logger.debug("DAR: {}".format(dar_uuid))

            if dar_uuid is None:
                logger.error("Unable to lookup address: {}".format(address_string))
                self.address_errors[unit_id] = info
                return

            self.importer.add_address_type(
                organisation_unit=unit_id,
                type_ref="AddressMailUnit",
                value=dar_uuid,
                date_from=date_from,
            )

        if self.create_email_addresses:
            import_email_addresses(info)
        import_pnumber(info)
        import_postal_address(info)

        # Recursively create parents (higher level OUs)
        # Note: These do not have their own entry in SD
        if "DepartmentReference" in department:
            self._add_sd_department(
                department["DepartmentReference"],
                contains_subunits=True,
                sub_tree=sub_tree,
                super_unit=super_unit,
            )

    def _get_institution(self) -> uuid.UUID:
        params = {
            "InstitutionIdentifier": self.settings.sd_institution_identifier,
            "UUIDIndicator": "true",
        }
        r = sd_lookup("GetInstitution20111201", settings=self.settings, params=params)
        return uuid.UUID(r["Region"]["Institution"]["InstitutionUUIDIdentifier"])

    def _create_org_tree_structure(self):
        nodes = {}
        all_ous = self.importer.export("organisation_unit")
        new_ous = []
        for key, ou in all_ous.items():
            parent = ou.parent_ref
            if parent is None:
                uuid = key
                niveau = ou.org_unit_level_ref
                nodes[uuid] = Node(niveau, uuid=uuid)
            else:
                new_ous.append(ou)

        while len(new_ous) > 0:
            logger.info("Number of new ous: {}".format(len(new_ous)))
            print("Number of new ous: {}".format(len(new_ous)))
            all_ous = new_ous
            new_ous = []
            for ou in all_ous:
                parent = ou.parent_ref
                if parent in nodes.keys():
                    uuid = ou.uuid
                    niveau = ou.org_unit_level_ref
                    nodes[uuid] = Node(niveau, parent=nodes[parent], uuid=uuid)
                else:
                    new_ous.append(ou)
        return nodes

    def _ad_creation_trigger_it_system(self, person: OrderedDict) -> None:
        """
        Some municipalitites wish to flag certain users in SD with a flag
        (a certain string) in an SD persons <TelephoneNumberIdentifier> in the
        <ContactInformation> in the <Employment> tag for the person. If the
        string is found, an IT-system is created on the user in MO, and this
        IT-system will in turn trigger the LDAP-integration to create the user
        in the AD.

        (see # See https://redmine.magenta-aps.dk/issues/56089)

        Args:
            person: The SD person to create the IT-system for
        """
        employment = person.get("Employment", [])
        employments = ensure_list(employment)
        for emp in employments:
            contact_info = emp.get("ContactInformation", {})
            telephone_number_ids = ensure_list(
                contact_info.get("TelephoneNumberIdentifier")
            )
            telephone_number_ids = [
                tni.strip() for tni in telephone_number_ids if tni is not None
            ]

            if self.settings.sd_phone_number_id_trigger in telephone_number_ids:
                self.importer.join_itsystem(
                    employee=person["PersonCivilRegistrationIdentifier"],
                    user_key=emp["EmploymentIdentifier"],
                    itsystem_ref=self.settings.sd_phone_number_id_for_ad_string,
                    date_from=format_date(self.settings.sd_global_from_date),
                )

    def add_people(self):
        """Load all person details and store for later user"""
        params = {
            "StatusActiveIndicator": "true",
            "StatusPassiveIndicator": "false",
            "ContactInformationIndicator": str(
                self.settings.sd_phone_number_id_for_ad_creation
            ).lower(),
            "PostalAddressIndicator": "false",
            "EffectiveDate": self.import_date,
        }
        active_people = sd_lookup(
            "GetPerson20111201", settings=self.settings, params=params
        )
        if not isinstance(active_people["Person"], list):
            active_people["Person"] = [active_people["Person"]]

        params["StatusActiveIndicator"] = False
        params["StatusPassiveIndicator"] = True
        passive_people = sd_lookup(
            "GetPerson20111201", settings=self.settings, params=params
        )
        if not isinstance(passive_people["Person"], list):
            passive_people["Person"] = [passive_people["Person"]]

        cprs = set(
            map(
                itemgetter("PersonCivilRegistrationIdentifier"), active_people["Person"]
            )
        )
        # Collect all people, prefering their active variants
        # TODO: Consider doing this with a dict keyed by cpr number
        people = active_people["Person"]
        for person in passive_people["Person"]:
            cpr = person["PersonCivilRegistrationIdentifier"]
            if cpr not in cprs:
                people.append(person)

        people = filter(skip_fictional_users, people)

        # TODO: Almost identitcal code exists in sd_changed_at's update_changed_persons
        for person in people:
            cpr = person["PersonCivilRegistrationIdentifier"]
            self._update_ad_map(cpr)

            given_name = person.get("PersonGivenName", "")
            sur_name = person.get("PersonSurnameName", "")

            uuid = self.employee_forced_uuids.get(cpr)
            logger.info(f"Employee in force list: {uuid}")
            if uuid is None and "ObjectGuid" in self.ad_people[cpr]:
                uuid = self.ad_people[cpr]["ObjectGuid"]

            # Name is placeholder for initals, do not know which field to extract
            if "Name" in self.ad_people[cpr]:
                user_key = self.ad_people[cpr]["Name"]
            else:
                user_key = "{} {}".format(given_name, sur_name)

            self.importer.add_employee(
                name=(given_name, sur_name),
                identifier=cpr,
                cpr_no=cpr,
                user_key=user_key,
                uuid=uuid,
            )

            if "SamAccountName" in self.ad_people[cpr]:
                self.importer.join_itsystem(
                    employee=cpr,
                    user_key=self.ad_people[cpr]["SamAccountName"],
                    itsystem_ref="AD",
                    date_from="1930-01-01",
                )

            # See https://redmine.magenta-aps.dk/issues/56089
            if self.settings.sd_phone_number_id_for_ad_creation:
                self._ad_creation_trigger_it_system(person)

    def create_ou_tree(self, create_orphan_container, sub_tree=None, super_unit=None):
        """Read all department levels from SD."""
        # TODO: Currently we can only read a top sub-tree
        if create_orphan_container:
            self._add_klasse("Orphan", "Virtuel Enhed", "org_unit_type")
            self.importer.add_organisation_unit(
                identifier="OrphanUnits",
                uuid="11111111-0000-0000-0000-111111111111",
                name="Forældreløse enheder",
                user_key="OrphanUnits",
                type_ref="Orphan",
                date_from="1930-01-01",
                date_to=None,
                parent_ref=None,
            )

        # Create historic dummy org unit
        historic_org_unit_date_to = format_date(
            date_to_datetime(self.settings.sd_global_from_date)
            - datetime.timedelta(days=1)
        )
        self._add_klasse(HISTORIC, "Historisk Enhed", "org_unit_type")
        self.importer.add_organisation_unit(
            identifier=self.historic_org_unit_uuid,
            uuid=self.historic_org_unit_uuid,
            name="Tidligere ansættelser",
            user_key=HISTORIC,
            type_ref=HISTORIC,
            date_from="1930-01-01",
            date_to=historic_org_unit_date_to,
            parent_ref=None,
        )

        params = {
            "ActivationDate": self.import_date,
            "DeactivationDate": self.import_date,
            "UUIDIndicator": "true",
        }
        organisation = sd_lookup(
            "GetOrganization20111201", settings=self.settings, params=params
        )
        departments = organisation["Organization"]["DepartmentReference"]

        for department in departments:
            self._add_sd_department(
                department, sub_tree=sub_tree, super_unit=super_unit
            )
        self.nodes = self._create_org_tree_structure()

    def create_employees(self):
        logger.info("Create employees")

        effective_date = datetime.datetime.strptime(self.import_date, "%d.%m.%Y").date()

        logger.info("Get active people from SD...")
        active_people = ensure_list(
            read_employment_at(
                effective_date,
                settings=self.settings,
                status_passive_indicator=False,
            )
        )

        logger.info("Get passive people from SD...")
        passive_people = ensure_list(
            read_employment_at(
                effective_date,
                settings=self.settings,
                status_active_indicator=False,
                status_passive_indicator=True,
            )
        )

        # TODO: condense the two calls above into one and then create the variables
        # active_people and passive_people by using the partition function from
        # more_itertools

        logger.info("Create employees from active SD employees...")
        self._create_employees(active_people)

        logger.info("Create employees from passive SD employees...")
        self._create_employees(passive_people)

    def _create_employees(self, persons):
        people = filter(skip_fictional_users, persons)
        for person in people:
            self.create_employee(person)

    def _get_employee_target_unit_uuid(
        self, too_deep: list[str], original_unit_uuid: uuid.UUID
    ) -> uuid.UUID:
        """
        The employees in SD are located in the units which have org_unit_level
        set to "Afdelings-niveau" for SD salary technical reasons (details
        unknown to us). In MO, however, the employees must be moved to a proper
        "NY-niveau" located higher up in the OU tree (e.g. necessary for the
        export to FK-org). This function calculates the MO target unit based
        on the argument "too deep" (i.e. the org_unit_levels in which employees
        are not allowed) which is typically just set to ["Afdelings-niveau"],
        but in some municipalities ["Afdelings-niveau", "NY1-niveau"] are also
        used.

        Args:
            too_deep: the org_unit_levels in which employees are not allowed
            original_unit_uuid: the SD unit where the employee is located

        Returns:
            The MO target unit for the employee
        """

        unit_uuid = str(original_unit_uuid)

        while self.nodes[unit_uuid].name in too_deep:
            if self.nodes[unit_uuid].parent is None:
                unit_uuid = str(original_unit_uuid)
                break
            else:
                unit_uuid = self.nodes[unit_uuid].parent.uuid

        return uuid.UUID(unit_uuid)

    def create_employee(self, person):
        logger.debug(79 * "-")
        logger.debug("Person object to create: {}".format(person))

        cpr = person["PersonCivilRegistrationIdentifier"]
        employments = ensure_list(person["Employment"])

        for employment in employments:
            status = EmploymentStatus(
                employment["EmploymentStatus"]["EmploymentStatusCode"]
            )
            if status in EmploymentStatus.let_go():
                continue

            # Job_position_id: Klassificeret liste over stillingstyper.
            # job_name: Fritekstfelt med stillingsbetegnelser.
            job_position_id = employment["Profession"]["JobPositionIdentifier"]
            if job_position_id in self.skip_job_functions:
                continue
            job_name = employment["Profession"].get("EmploymentName", job_position_id)

            occupation_rate = float(employment["WorkingTime"]["OccupationRate"])

            employment_id = calc_employment_id(employment)

            # TODO: Identital code to this exists in sd_changed_at
            split = self.settings.sd_monthly_hourly_divide
            if employment_id["value"] < split:
                engagement_type_ref = "månedsløn"
            elif (split - 1) < employment_id["value"] < 999999:
                engagement_type_ref = "timeløn"
            else:  # This happens if EmploymentID is not a number
                engagement_type_ref = "engagement_type" + job_position_id
                self._add_klasse(
                    engagement_type_ref, job_position_id, "engagement_type"
                )
                logger.info("Non-numeric id. Job pos id: {}".format(job_position_id))

            primary_type_ref = "non-primary"
            # If status 0, uncondtionally override
            if status == EmploymentStatus.AnsatUdenLoen:
                primary_type_ref = "status0"

            job_function_type = self.settings.sd_job_function
            if job_function_type == JobFunction.employment_name:
                job_func_ref = self._add_klasse(
                    job_name, job_name, "engagement_job_function"
                )
            elif job_function_type == JobFunction.job_position_identifier:
                job_func_ref = self._add_klasse(
                    job_position_id, job_position_id, "engagement_job_function"
                )

            emp_dep = employment["EmploymentDepartment"]
            unit = emp_dep["DepartmentUUIDIdentifier"]

            datetime_from_engagement, datetime_to = get_employment_datetimes(employment)
            # Use org unit start date if engagement starts *before*
            # the org unit start date
            org_unit_datetime_from = parse_datetime(
                self.importer.organisation_units[unit].date_from
            )
            datetime_from = max(datetime_from_engagement, org_unit_datetime_from)

            date_from_str = format_date(datetime_from)
            date_to_str = format_date(datetime_to)
            if datetime_to == datetime.datetime(9999, 12, 31, 0, 0):
                date_to_str = None

            sd_employment_id = employment_id["id"]
            logger.info(
                f"Validty for {sd_employment_id}: from: {date_from_str}, "
                f"to: {date_to_str}"
            )

            if datetime_to <= datetime_from:
                logger.warning(f"Skip creating employment for id: {sd_employment_id}")
                continue

            original_unit = unit
            # Remove this to remove any sign of the employee from the
            # lowest levels of the org
            if self.create_associations:
                self.importer.add_association(
                    employee=cpr,
                    user_key=employment_id["id"],
                    organisation_unit=original_unit,
                    association_type_ref="SD-medarbejder",
                    date_from=date_from_str,
                    date_to=date_to_str,
                )

            too_deep = self.settings.sd_import_too_deep
            unit = str(
                self._get_employee_target_unit_uuid(too_deep, uuid.UUID(original_unit))
            )

            ext_field = self.settings.sd_employment_field
            extention = {}
            if ext_field is not None:
                extention[ext_field] = job_name

            # Generate UUID for engagement here since we need this UUID when/if
            # creating a leave below
            engagement_uuid = str(uuid.uuid4())

            self.importer.add_engagement(
                employee=cpr,
                uuid=engagement_uuid,
                user_key=employment_id["id"],
                organisation_unit=unit,
                job_function_ref=job_func_ref,
                fraction=int(occupation_rate * 1000000),
                primary_ref=primary_type_ref,
                engagement_type_ref=engagement_type_ref,
                date_from=date_from_str,
                date_to=date_to_str,
                **extention,
            )

            # Add historic dummy engagement if the start date of the engagement
            # is older than the start date of the corresponding org unit
            # (see https://redmine.magenta-aps.dk/issues/51898)
            if datetime_from_engagement < org_unit_datetime_from:
                dummy_eng_date_to = org_unit_datetime_from - datetime.timedelta(days=1)
                dummy_eng_date_to_str = format_date(dummy_eng_date_to)
                self.importer.add_engagement(
                    employee=cpr,
                    organisation_unit=self.historic_org_unit_uuid,
                    job_function_ref=HISTORIC,
                    engagement_type_ref="historisk",
                    date_from=format_date(datetime_from_engagement),
                    date_to=dummy_eng_date_to_str,
                )

            if status == EmploymentStatus.Orlov:
                self.importer.add_leave(
                    employee=cpr,
                    engagement_uuid=engagement_uuid,
                    leave_type_ref="Orlov",
                    date_from=date_from_str,
                    date_to=date_to_str,
                )

            # These job functions will normally (but necessarily)
            #  correlate to a manager position
            if job_position_id in ["1040", "1035", "1030"]:
                self.importer.add_manager(
                    employee=cpr,
                    organisation_unit=unit,
                    manager_level_ref="manager_" + job_position_id,
                    address_uuid=None,  # Manager address is not used
                    manager_type_ref="leder_type",
                    responsibility_list=["Lederansvar"],
                    date_from=date_from_str,
                    date_to=employment["Profession"]["DeactivationDate"],
                )


@click.group()
def cli():
    for name in logging.root.manager.loggerDict:
        if name in (
            "sdImport",
            "sdCommon",
            "AdReader",
            "moImporterMoraTypes",
            "moImporterMoxTypes",
            "moImporterUtilities",
            "moImporterHelpers",
        ):
            logging.getLogger(name).setLevel(LOG_LEVEL)
        else:
            logging.getLogger(name).setLevel(logging.ERROR)

    logging.basicConfig(
        format="%(levelname)s %(asctime)s %(name)s %(message)s",
        level=LOG_LEVEL,
        stream=sys.stdout,
    )


@cli.command()
@click.option(
    "--org-only",
    is_flag=True,
    default=False,
    type=click.BOOL,
    help="Only import organisation structure",
)
@click.option(
    "--mora-base",
    required=False,
    help="URL for OS2mo.",
)
@click.option(
    "--mox-base",
    required=False,
    help="URL for LoRa.",
)
def full_import_cli(org_only: bool, mora_base: Optional[str], mox_base: Optional[str]):
    """Tool to do an initial full import."""
    full_import(org_only, mora_base, mox_base)


def full_import(
    org_only: bool = False,
    mora_base: Optional[str] = None,
    mox_base: Optional[str] = None,
):
    """Tool to do an initial full import."""
    overrides = {}
    if mora_base:
        overrides["mora_base"] = mora_base
    if mox_base:
        overrides["mox_base"] = mox_base

    settings = get_importer_settings(**overrides)

    # Check connection to MO before we fire requests against SD
    mh = MoraHelper(settings.mora_base)
    if not mh.check_connection():
        raise click.ClickException("No MO reply, aborting.")

    importer = ImportHelper(
        create_defaults=True,
        mox_base=settings.mox_base,
        mora_base=settings.mora_base,
        seperate_names=True,
    )
    sd = SdImport(importer, settings=settings, org_only=org_only, ad_info=None)

    sd.create_ou_tree(create_orphan_container=False, sub_tree=None, super_unit=None)
    if not org_only:
        sd.create_employees()

    importer.import_all()
    print("IMPORT DONE")


if __name__ == "__main__":
    cli()



from raclients.graph.client import GraphQLClient


def get_mo_client(
    auth_server: str,
    client_id: str,
    client_secret: str,
    mo_base_url: str,
    gql_version: int,
    timeout: int = 600,
) -> GraphQLClient:
    """
    Get the GraphQL client for calling MO.

    Args:
        auth_server: the Keycloak server
        client_id: Keycloak client ID
        client_secret: Keycloak client secret
        mo_base_url: MOs base URL
        gql_version: GraphQL version
        timeout: timeout for the client

    Returns:
        A GraphQL client
    """

    return GraphQLClient(
        url=f"{mo_base_url}/graphql/v{str(gql_version)}",
        client_id=client_id,
        client_secret=client_secret,
        auth_server=auth_server,
        auth_realm="mo",
        execute_timeout=timeout,
        httpx_client_kwargs={"timeout": timeout},
        sync=True,
    )
